import json

import pytest

from ape.managers._deploymentscache import DeploymentDiskCache


class TestDeploymentDiskCache:
    CONTRACT_NAME = "DeploymentTestContractName"

    @pytest.fixture(scope="class")
    def contract_name(self):
        return self.CONTRACT_NAME

    @pytest.fixture
    def cache(self):
        return DeploymentDiskCache()

    def test_cache_deployment(self, zero_address, cache, contract_name):
        cache.cache_deployment(zero_address, contract_name)
        assert contract_name in cache
        assert cache[contract_name][-1].address == zero_address

    def test_cache_deployment_live_network(
        self, zero_address, cache, contract_name, mock_sepolia, eth_tester_provider
    ):
        local = eth_tester_provider.network
        ecosystem_name = mock_sepolia.ecosystem.name

        eth_tester_provider.network = mock_sepolia
        cache.cache_deployment(zero_address, contract_name)
        eth_tester_provider.network = local

        assert contract_name in cache
        assert cache[contract_name][-1].address == zero_address
        # Show it is also cached on disk.
        disk_data = json.loads(cache.cachefile.read_text())
        assert (
            disk_data["ecosystems"][ecosystem_name][mock_sepolia.name][contract_name][0]["address"]
            == zero_address
        )

    def test_cache_deployment_live_network_new_ecosystem(
        self, zero_address, cache, contract_name, mock_sepolia, eth_tester_provider
    ):
        """
        Tests the case when caching a deployment in a new ecosystem.
        """
        ecosystem_name = mock_sepolia.ecosystem.name
        local = eth_tester_provider.network
        eth_tester_provider.network = mock_sepolia
        # Make the ecosystem key not exist.
        deployments = cache._deployments.pop(ecosystem_name, None)
        cache.cache_deployment(zero_address, contract_name)
        eth_tester_provider.network = local
        if deployments is not None:
            cache._deployments[ecosystem_name] = deployments
        cache.cachefile.unlink(missing_ok=True)

        # In memory cached still work.
        assert contract_name in cache
        assert cache[contract_name][-1].address == zero_address

        # Show it did NOT cache to disk.
        if cache.cachefile.is_file():
            disk_data = json.loads(cache.cachefile.read_text())
            assert contract_name not in disk_data["ecosystems"][ecosystem_name]["sepolia"]



from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from ape_ethereum.ecosystem import EthereumConfig

    return EthereumConfig


@plugins.register(plugins.ConversionPlugin)
def converters():
    from ape_ethereum._converters import WeiConversions

    yield int, WeiConversions


@plugins.register(plugins.EcosystemPlugin)
def ecosystems():
    from ape_ethereum.ecosystem import Ethereum

    yield Ethereum


@plugins.register(plugins.NetworkPlugin)
def networks():
    from ape.api.networks import ForkedNetworkAPI, NetworkAPI, create_network_type
    from ape_ethereum.ecosystem import NETWORKS

    for network_name, network_params in NETWORKS.items():
        yield "ethereum", network_name, create_network_type(*network_params)
        yield "ethereum", f"{network_name}-fork", ForkedNetworkAPI

    # NOTE: This works for local providers, as they get chain_id from themselves
    yield "ethereum", "local", NetworkAPI


@plugins.register(plugins.QueryPlugin)
def query_engines():
    from .query import EthereumQueryProvider

    yield EthereumQueryProvider


def __getattr__(name):
    if name in (
        "BaseEthereumConfig",
        "Block",
        "Ethereum",
        "EthereumConfig",
        "ForkedNetworkConfig",
        "NetworkConfig",
    ):
        import ape_ethereum.ecosystem as ecosystem_module

        return getattr(ecosystem_module, name)

    elif name in (
        "EthereumNodeProvider",
        "Web3Provider",
        "assert_web3_provider_uri_env_var_not_set",
    ):
        import ape_ethereum.provider as provider_module

        return getattr(provider_module, name)

    elif name in (
        "AccessListTransaction",
        "Authorization",
        "BaseTransaction",
        "DynamicFeeTransaction",
        "Receipt",
        "SetCodeTransaction",
        "SharedBlobReceipt",
        "SharedBlobTransaction",
        "StaticFeeTransaction",
        "TransactionStatusEnum",
        "TransactionType",
    ):
        import ape_ethereum.transactions as tx_module

        return getattr(tx_module, name)

    elif name in ("CallTrace", "Trace", "TransactionTrace"):
        import ape_ethereum.trace as trace_module

        return getattr(trace_module, name)

    else:
        raise AttributeError(name)


__all__ = [
    "AccessListTransaction",
    "Authorization",
    "BaseEthereumConfig",
    "BaseTransaction",
    "Block",
    "CallTrace",
    "DynamicFeeTransaction",
    "Ethereum",
    "EthereumConfig",
    "EthereumNodeProvider",
    "ForkedNetworkConfig",
    "NetworkConfig",
    "Receipt",
    "SetCodeTransaction",
    "SharedBlobReceipt",
    "SharedBlobTransaction",
    "SharedBlobTransaction",
    "StaticFeeTransaction",
    "Trace",
    "TransactionStatusEnum",
    "TransactionTrace",
    "TransactionType",
    "Web3Provider",
    "assert_web3_provider_uri_env_var_not_set",
]



def __getattr__(name: str):
    if name == "abstractmethod":
        from abc import abstractmethod

        return abstractmethod

    elif name in (
        "LogInputABICollection",
        "Struct",
        "StructParser",
        "is_array",
        "is_dynamic_sized_type",
        "is_named_tuple",
        "is_struct",
        "returns_array",
    ):
        import ape.utils.abi as abi_module

        return getattr(abi_module, name)

    elif name in (
        "BaseInterface",
        "BaseInterfaceModel",
        "ExtraAttributesMixin",
        "ExtraModelAttributes",
        "ManagerAccessMixin",
        "injected_before_use",
        "only_raise_attribute_error",
    ):
        import ape.utils.basemodel as basemodel_module

        return getattr(basemodel_module, name)

    elif name in (
        "DEFAULT_LIVE_NETWORK_BASE_FEE_MULTIPLIER",
        "DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT",
        "DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT",
        "EMPTY_BYTES32",
        "LOCAL_NETWORK_NAME",
        "SOURCE_EXCLUDE_PATTERNS",
        "ZERO_ADDRESS",
        "add_padding_to_strings",
        "as_our_module",
        "cached_property",
        "extract_nested_value",
        "gas_estimation_error_message",
        "get_current_timestamp_ms",
        "get_package_version",
        "is_evm_precompile",
        "is_zero_hex",
        "load_config",
        "log_instead_of_fail",
        "nonreentrant",
        "pragma_str_to_specifier_set",
        "raises_not_implemented",
        "run_until_complete",
        "singledispatchmethod",
        "to_int",
    ):
        import ape.utils.misc as misc_module

        return getattr(misc_module, name)

    elif name in (
        "clean_path",
        "create_tempdir",
        "expand_environment_variables",
        "extract_archive",
        "get_all_files_in_directory",
        "get_full_extension",
        "get_package_path",
        "get_relative_path",
        "in_tempdir",
        "path_match",
        "run_in_tempdir",
        "use_temp_sys_path",
    ):
        import ape.utils.os as os_module

        return getattr(os_module, name)

    elif name in ("JoinableQueue", "spawn"):
        import ape.utils.process as process_module

        return getattr(process_module, name)

    elif name in (
        "USER_AGENT",
        "RPCHeaders",
        "allow_disconnected",
        "request_with_retry",
        "stream_response",
    ):
        import ape.utils.rpc as rpc_module

        return getattr(rpc_module, name)

    elif name in (
        "DEFAULT_NUMBER_OF_TEST_ACCOUNTS",
        "DEFAULT_TEST_ACCOUNT_BALANCE",
        "DEFAULT_TEST_CHAIN_ID",
        "DEFAULT_TEST_HD_PATH",
        "DEFAULT_TEST_MNEMONIC",
        "GeneratedDevAccount",
        "generate_dev_accounts",
    ):
        import ape.utils.testing as testing_module

        return getattr(testing_module, name)

    elif name in (
        "USER_ASSERT_TAG",
        "TraceStyles",
        "parse_coverage_tables",
        "parse_gas_table",
        "prettify_dict",
        "prettify_function",
        "prettify_inputs",
        "prettify_list",
    ):
        import ape.utils.trace as trace_module

        return getattr(trace_module, name)

    else:
        raise AttributeError(name)


__all__ = [
    "DEFAULT_LIVE_NETWORK_BASE_FEE_MULTIPLIER",
    "DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT",
    "DEFAULT_NUMBER_OF_TEST_ACCOUNTS",
    "DEFAULT_TEST_ACCOUNT_BALANCE",
    "DEFAULT_TEST_CHAIN_ID",
    "DEFAULT_TEST_HD_PATH",
    "DEFAULT_TEST_MNEMONIC",
    "DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT",
    "EMPTY_BYTES32",
    "LOCAL_NETWORK_NAME",
    "SOURCE_EXCLUDE_PATTERNS",
    "USER_AGENT",
    "USER_ASSERT_TAG",
    "ZERO_ADDRESS",
    "BaseInterface",
    "BaseInterfaceModel",
    "ExtraAttributesMixin",
    "ExtraModelAttributes",
    "GeneratedDevAccount",
    "JoinableQueue",
    "LogInputABICollection",
    "ManagerAccessMixin",
    "RPCHeaders",
    "Struct",
    "StructParser",
    "TraceStyles",
    "abstractmethod",
    "add_padding_to_strings",
    "allow_disconnected",
    "as_our_module",
    "cached_property",
    "clean_path",
    "create_tempdir",
    "expand_environment_variables",
    "extract_archive",
    "extract_nested_value",
    "gas_estimation_error_message",
    "generate_dev_accounts",
    "get_all_files_in_directory",
    "get_current_timestamp_ms",
    "get_full_extension",
    "get_package_path",
    "get_package_version",
    "get_relative_path",
    "in_tempdir",
    "injected_before_use",
    "is_array",
    "is_dynamic_sized_type",
    "is_evm_precompile",
    "is_named_tuple",
    "is_struct",
    "is_zero_hex",
    "load_config",
    "log_instead_of_fail",
    "nonreentrant",
    "only_raise_attribute_error",
    "parse_coverage_tables",
    "parse_gas_table",
    "path_match",
    "prettify_dict",
    "prettify_function",
    "prettify_inputs",
    "prettify_list",
    "pragma_str_to_specifier_set",
    "raises_not_implemented",
    "request_with_retry",
    "returns_array",
    "run_in_tempdir",
    "run_until_complete",
    "singledispatchmethod",
    "spawn",
    "stream_response",
    "to_int",
    "use_temp_sys_path",
]



import atexit
import os.path
import re
import shutil
from pathlib import Path
from subprocess import DEVNULL, PIPE, Popen
from typing import TYPE_CHECKING, Any
from urllib.error import HTTPError, URLError
from urllib.parse import urlparse
from urllib.request import urlopen

from eth_utils import add_0x_prefix, to_hex
from geth.chain import initialize_chain as initialize_gethdev_chain
from geth.process import BaseGethProcess
from geth.wrapper import ALL_APIS, construct_test_chain_kwargs
from pydantic import field_validator
from pydantic_settings import SettingsConfigDict
from requests.exceptions import ConnectionError
from web3.middleware import ExtraDataToPOAMiddleware

from ape.api.config import PluginConfig
from ape.api.providers import SubprocessProvider, TestProviderAPI
from ape.exceptions import VirtualMachineError
from ape.logging import LogLevel, logger
from ape.utils.misc import ZERO_ADDRESS, log_instead_of_fail, raises_not_implemented
from ape.utils.process import JoinableQueue, spawn
from ape.utils.testing import (
    DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    DEFAULT_TEST_ACCOUNT_BALANCE,
    DEFAULT_TEST_CHAIN_ID,
    DEFAULT_TEST_HD_PATH,
    DEFAULT_TEST_MNEMONIC,
    generate_dev_accounts,
)
from ape_ethereum.provider import (
    DEFAULT_HOSTNAME,
    DEFAULT_PORT,
    DEFAULT_SETTINGS,
    EthereumNodeProvider,
)
from ape_ethereum.trace import TraceApproach

if TYPE_CHECKING:
    from geth.types import GenesisDataTypedDict

    from ape.api.accounts import TestAccountAPI
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.types.vm import SnapshotID


Alloc = dict[str, dict[str, Any]]


def create_genesis_data(alloc: Alloc, chain_id: int) -> "GenesisDataTypedDict":
    """
    A wrapper around genesis data for py-geth that
    fills in more defaults.
    """
    return {
        "alloc": alloc,
        "config": {
            "arrowGlacierBlock": 0,
            "berlinBlock": 0,
            "byzantiumBlock": 0,
            "cancunTime": 0,
            "chainId": chain_id,
            "constantinopleBlock": 0,
            "daoForkBlock": 0,
            "daoForkSupport": True,
            "eip150Block": 0,
            "eip155Block": 0,
            "eip158Block": 0,
            "ethash": {},
            "grayGlacierBlock": 0,
            "homesteadBlock": 0,
            "istanbulBlock": 0,
            "londonBlock": 0,
            "petersburgBlock": 0,
            "shanghaiTime": 0,
            "terminalTotalDifficulty": 0,
            "terminalTotalDifficultyPassed": True,
            "blobSchedule": {
                "cancun": {"target": 3, "max": 6, "baseFeeUpdateFraction": 3338477},
                "prague": {"target": 6, "max": 9, "baseFeeUpdateFraction": 5007716},
            },
        },
        "coinbase": ZERO_ADDRESS,
        "difficulty": "0x0",
        "gasLimit": "0x0",
        "extraData": "0x",
        "mixHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
        "nonce": "0x0",
        "timestamp": "0x0",
        "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
        "baseFeePerGas": "0x0",
    }


def get_node_name_from_executable(executable: str) -> str:
    return executable.split(os.path.sep)[-1] or "geth"


class GethDevProcess(BaseGethProcess):
    """
    A developer-configured geth that only exists until disconnected.
    (Implementation detail of the local node provider).
    """

    def __init__(
        self,
        data_dir: Path,
        hostname: str | None = None,
        port: int | None = None,
        ipc_path: Path | None = None,
        ws_hostname: str | None = None,
        ws_port: str | None = None,
        mnemonic: str = DEFAULT_TEST_MNEMONIC,
        number_of_accounts: int = DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
        chain_id: int = DEFAULT_TEST_CHAIN_ID,
        initial_balance: str | int = DEFAULT_TEST_ACCOUNT_BALANCE,
        executable: list[str] | str | None = None,
        auto_disconnect: bool = True,
        extra_funded_accounts: list[str] | None = None,
        hd_path: str | None = DEFAULT_TEST_HD_PATH,
        block_time: int | None = None,
        generate_accounts: bool = True,
        initialize_chain: bool = True,
        background: bool = False,
        verify_bin: bool = True,
        rpc_api: list[str] | None = None,
    ):
        if isinstance(executable, str):
            # Legacy.
            executable = [executable]

        if executable:
            if verify_bin and (
                not Path(executable[0]).exists() and not shutil.which(executable[0])
            ):
                raise NodeSoftwareNotInstalledError()

        else:
            # Executable not specified: attempt to find one in $PATH.
            if shutil.which("geth"):
                executable = ["geth"]
            elif shutil.which("reth"):
                executable = ["reth", "node"]
            elif verify_bin:
                # TODO: Add support for more nodes, such as erigon.
                raise NodeSoftwareNotInstalledError()
            else:
                # This probably won't work if started, but the user knows that.
                executable = ["geth"]

        self.executable = executable or ["geth"]
        self._data_dir = data_dir
        self.is_running = False
        self._auto_disconnect = auto_disconnect
        self.background = background

        is_reth = executable[0].endswith("reth")

        if is_reth and hostname == "localhost":
            # Reth only supports IP format.
            hostname = "127.0.0.1"

        kwargs_ctor: dict = {
            "data_dir": self.data_dir,
            "geth_executable": executable[0],
        }
        if hostname is not None:
            kwargs_ctor["rpc_addr"] = hostname
        if port is not None:
            kwargs_ctor["rpc_port"] = f"{port}"
        if ws_hostname:
            kwargs_ctor["ws_enabled"] = True
            kwargs_ctor["ws_addr"] = ws_hostname
        if ws_port:
            kwargs_ctor["ws_enabled"] = True
            kwargs_ctor["ws_port"] = f"{ws_port}"
        if ipc_path is not None:
            kwargs_ctor["ipc_path"] = f"{ipc_path}"
        if not kwargs_ctor.get("ws_enabled"):
            kwargs_ctor["ws_api"] = None
            kwargs_ctor["ws_enabled"] = False
            kwargs_ctor["ws_addr"] = None
            kwargs_ctor["ws_port"] = None
        if block_time is not None and not is_reth:
            kwargs_ctor["dev_period"] = f"{block_time}"

        if rpc_api is None:
            # Reth also has MEV API support.
            rpc_api_str = f"{ALL_APIS},mev" if is_reth else ALL_APIS
        else:
            rpc_api_str = ",".join(rpc_api)

        kwargs_ctor["rpc_api"] = rpc_api_str
        geth_kwargs = construct_test_chain_kwargs(**kwargs_ctor)
        if is_reth:
            geth_kwargs.pop("max_peers", None)
            geth_kwargs.pop("no_discover", None)

            # NOTE: --verbosity _is_ available in reth, but it is a different type (flag only).
            #   It isn't really needed anyway (I don't think).
            geth_kwargs.pop("verbosity", None)

            geth_kwargs.pop("password", None)

        # Ensure IPC path has correct name.
        if ipc_path_kwarg := geth_kwargs.get("ipc_path"):
            if ipc_path_kwarg.endswith("geth.ipc"):
                geth_kwargs["ipc_path"] = ipc_path_kwarg.replace(
                    "geth.ipc", f"{self.process_name}.ipc"
                )

        # Ensure a clean data-dir.
        self._clean()

        geth_kwargs["dev_mode"] = True

        # Does not work with --dev mode.
        geth_kwargs.pop("network_id", None)

        hd_path = hd_path or DEFAULT_TEST_HD_PATH

        if generate_accounts:
            self._dev_accounts = generate_dev_accounts(
                mnemonic, number_of_accounts=number_of_accounts, hd_path=hd_path
            )
        else:
            self._dev_accounts = []

        if initialize_chain:
            addresses = [a.address for a in self._dev_accounts]
            addresses.extend(extra_funded_accounts or [])
            bal_dict = {"balance": str(initial_balance)}
            alloc = dict.fromkeys(addresses, bal_dict)
            genesis = create_genesis_data(alloc, chain_id)
            self._data_dir.mkdir(parents=True, exist_ok=True)
            initialize_gethdev_chain(genesis, self.data_dir)

        super().__init__(geth_kwargs)

        # Correct multi-word executable.
        idx = self.command.index(executable[0])
        self.command = self.command[:idx] + executable + self.command[idx + 1 :]

    @classmethod
    def from_uri(cls, uri: str, data_folder: Path, **kwargs):
        mnemonic = kwargs.get("mnemonic", DEFAULT_TEST_MNEMONIC)
        number_of_accounts = kwargs.get("number_of_accounts", DEFAULT_NUMBER_OF_TEST_ACCOUNTS)
        balance = kwargs.get("initial_balance", DEFAULT_TEST_ACCOUNT_BALANCE)
        extra_accounts = [a.lower() for a in kwargs.get("extra_funded_accounts", [])]
        block_time = kwargs.get("block_time", None)
        if isinstance(block_time, int):
            block_time = f"{block_time}"

        executable = kwargs.get("executable")

        process_kwargs = {
            "auto_disconnect": kwargs.get("auto_disconnect", True),
            "background": kwargs.get("background", False),
            "block_time": block_time,
            "executable": executable,
            "extra_funded_accounts": extra_accounts,
            "hd_path": kwargs.get("hd_path", DEFAULT_TEST_HD_PATH),
            "initial_balance": balance,
            "mnemonic": mnemonic,
            "number_of_accounts": number_of_accounts,
            "verify_bin": kwargs.get("verify_bin", True),
        }

        parsed_uri = urlparse(uri)
        if not parsed_uri.netloc:
            path = Path(parsed_uri.path)
            if path.suffix == ".ipc":
                # Was given an IPC path.
                process_kwargs["ipc_path"] = path

            else:
                raise ConnectionError(f"Unrecognized path type: '{path}'.")

        elif hostname := parsed_uri.hostname:
            if hostname not in ("localhost", "127.0.0.1"):
                name = get_node_name_from_executable(executable) if executable else "geth"
                raise ConnectionError(
                    f"Unable to start {name} on non-local host {parsed_uri.hostname}."
                )

            if parsed_uri.scheme.startswith("ws"):
                process_kwargs["ws_hostname"] = hostname
                process_kwargs["ws_port"] = parsed_uri.port or DEFAULT_PORT
            elif parsed_uri.scheme.startswith("http"):
                process_kwargs["hostname"] = hostname or DEFAULT_HOSTNAME
                process_kwargs["port"] = parsed_uri.port or DEFAULT_PORT
            else:
                raise ConnectionError(f"Unsupported scheme: '{parsed_uri.scheme}'.")

        return cls(data_folder, **process_kwargs)

    @property
    def data_dir(self) -> str:
        return f"{self._data_dir}"

    @property
    def process_name(self) -> str:
        return get_node_name_from_executable(self.executable[0])

    @property
    def ipc_path(self) -> str:
        # Overridden: so we can use a custom IPC path name (e.g. reth.ipc).
        return self.geth_kwargs.get("ipc_path") or os.path.abspath(
            os.path.expanduser(
                os.path.join(
                    self.data_dir,
                    f"{self.process_name}.ipc",
                )
            )
        )

    @property
    def is_rpc_ready(self) -> bool:
        # Overridden: This is overridden to work with other nodes besides `geth`.
        #   Otherwise, the RPC is never declared as ready even though it is.
        try:
            urlopen(f"http://{self.rpc_host}:{self.rpc_port}")

        except HTTPError:
            # Reth nodes (and maybe others) might throw an HTTP error here, like "method not found".
            # This means the RPC is ready.
            return True

        except URLError:
            # Nothing found at all yet, most likely.
            return False

        else:
            # No error occurs on Geth nodes when the RPC is ready.
            return True

    @property
    def _hostname(self) -> str | None:
        return self.geth_kwargs.get("rpc_addr")

    @property
    def _port(self) -> str | None:
        return self.geth_kwargs.get("rpc_port")

    @property
    def _ws_hostname(self) -> str | None:
        return self.geth_kwargs.get("ws_addr")

    @property
    def _ws_port(self) -> str | None:
        return self.geth_kwargs.get("ws_port")

    def connect(self, timeout: int = 60):
        self._log_connection()
        self.start()
        self.wait_for_rpc(timeout=timeout)

        # Register atexit handler to make sure disconnect is called for normal object lifecycle.
        if self._auto_disconnect:
            atexit.register(self.disconnect)

    def _log_connection(self):
        home = str(Path.home())
        ipc_path = self.ipc_path.replace(home, "$HOME")

        http_log = ""
        if self._hostname:
            http_log = f"HTTP={self._hostname}"
            if port := self._port:
                http_log = f"{http_log}:{port}"

        ipc_log = f"IPC={ipc_path}"

        ws_log = ""
        if self._ws_hostname:
            ws_log = f"WS={self._ws_hostname}"
            if port := self._ws_port:
                ws_log = f"{ws_log}:{port}"

        connection_logs = ", ".join(x for x in (http_log, ipc_log, ws_log) if x)
        logger.info(f"Starting {self.process_name} ({connection_logs}).")

    def start(self):
        if self.is_running:
            return

        self.is_running = True

        if self.background or logger.level > LogLevel.DEBUG:
            out_file = DEVNULL
        else:
            out_file = PIPE

        self.proc = Popen(
            self.command,
            stdin=PIPE,
            stdout=out_file,
            stderr=out_file,
        )

    def disconnect(self):
        if self.is_running:
            logger.info(f"Stopping '{self.process_name}' process.")
            self.stop()

        self._clean()

    def _clean(self):
        if self._data_dir.is_dir():
            shutil.rmtree(self._data_dir)

    def wait(self, *args, **kwargs):
        if self.proc is None:
            return

        self.proc.wait(*args, **kwargs)


class EthereumNetworkConfig(PluginConfig):
    # Make sure you are running the right networks when you try for these
    mainnet: dict = {}
    holesky: dict = {}
    sepolia: dict = {}
    # Make sure to run via `geth --dev` (or similar)
    local: dict = {**DEFAULT_SETTINGS.copy(), "chain_id": DEFAULT_TEST_CHAIN_ID, "block_time": 0}

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NODE_")

    @field_validator("local", mode="before")
    @classmethod
    def _validate_local(cls, value):
        value = value or {}
        if not value:
            return {**DEFAULT_SETTINGS.copy(), "chain_id": DEFAULT_TEST_CHAIN_ID}

        if "chain_id" not in value:
            value["chain_id"] = DEFAULT_TEST_CHAIN_ID
        if ("uri" not in value and "ipc_path" in value) or "ws_uri" in value or "http_uri" in value:
            # No need to add default HTTP URI if was given only IPC Path
            return {**{k: v for k, v in DEFAULT_SETTINGS.items() if k != "uri"}, **value}

        return {**DEFAULT_SETTINGS, **value}


class EthereumNodeConfig(PluginConfig):
    """
    Configure your ``node:`` in Ape, the default provider
    plugin for live-network nodes. Also, ``ape node`` can
    start-up a local development node for testing purposes.
    """

    ethereum: EthereumNetworkConfig = EthereumNetworkConfig()
    """
    Configure the Ethereum network settings for the ``ape node`` provider,
    such as which URIs to use for each network.
    """

    executable: list[str] | None = None
    """
    For starting nodes, select the executable. Defaults to using
    ``shutil.which("geth")``.
    """

    data_dir: Path | None = None
    """
    For node-management, choose where the geth data directory shall
    be located. Defaults to using a location within Ape's DATA_FOLDER.
    """

    ipc_path: Path | None = None
    """
    For IPC connections, select the IPC path. If managing a process,
    web3.py can determine the IPC w/o needing to manually configure.
    """

    call_trace_approach: TraceApproach | None = None
    """
    Select the trace approach to use. Defaults to deducing one
    based on your node's client-version and available RPCs.
    """

    request_headers: dict = {}
    """
    Optionally specify request headers to use whenever using this provider.
    """

    rpc_api: list[str] | None = None
    """
    RPC APIs to enable. Defaults to all geth APIs.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NODE_")

    @field_validator("call_trace_approach", mode="before")
    @classmethod
    def validate_trace_approach(cls, value):
        # This handles nicer config values.
        return None if value is None else TraceApproach.from_key(value)

    @field_validator("executable", mode="before")
    @classmethod
    def validate_executable(cls, value):
        if not value:
            return None

        elif isinstance(value, str):
            return value.split(" ")

        return value


class NodeSoftwareNotInstalledError(ConnectionError):
    def __init__(self):
        super().__init__(
            "No node found and 'ape-node' is unable to start one.\n"
            "Things you can do:\n"
            "\t1. Check your connection URL, if trying to connect remotely.\n"
            "\t2. Install node software (geth), if trying to run a local node.\n"
            "\t3. Use and configure a different provider plugin, such as 'ape-foundry'."
        )


# NOTE: Using EthereumNodeProvider because of it's geth-derived default behavior.
# TODO: In 0.9, change NAME to be `gethdev`, so for local networks it is more obvious.
class GethDev(EthereumNodeProvider, TestProviderAPI, SubprocessProvider):
    _process: GethDevProcess | None = None
    name: str = "node"

    @property
    def process_name(self) -> str:
        if self._process:
            return self._process.process_name

        elif exec_cfg := self.config.executable:
            return get_node_name_from_executable(exec_cfg[0])

        return "geth"

    @property
    def chain_id(self) -> int:
        return self.settings.ethereum.local.get("chain_id", DEFAULT_TEST_CHAIN_ID)

    @property
    def block_time(self) -> int | None:
        return self.settings.ethereum.local.get("block_time")

    @property
    def data_dir(self) -> Path:
        # Overridden from base class for placing debug logs in ape data folder.
        return self.settings.data_dir or self.config_manager.DATA_FOLDER / self.name

    @log_instead_of_fail(default="<node>")
    def __repr__(self) -> str:
        client_version = self.client_version
        client_version_str = f" ({client_version}) " if client_version else " "
        return f"<Node{client_version_str}chain_id={self.chain_id}>"

    @property
    def auto_mine(self) -> bool:
        if self.process is not None:
            # Geth --dev auto mines.
            return True

        try:
            return self.make_request("eth_mining", [])
        except NotImplementedError:
            # Assume true; unlikely to be off. Geth --dev automines.
            return True

    @auto_mine.setter
    def auto_mine(self, value):
        raise NotImplementedError("'auto_mine' setter not implemented.")

    @property
    def ipc_path(self) -> Path | None:
        if rpc := self._configured_ipc_path:
            # "ipc_path" found in config/settings
            return Path(rpc)

        elif rpc := self._configured_uri:
            if f"{rpc}".endswith(".ipc"):
                # "uri" found in config/settings and is IPC.
                return Path(rpc)

        elif proc := self._process:
            # Connected.
            return Path(proc.ipc_path)

        # Default (used by geth-process).
        return self.data_dir / self.process_name / f"{self.process_name}.ipc"

    def connect(self):
        self._set_web3()
        if self.is_connected:
            self._complete_connect()

        elif self.allow_start:
            # Starting the process.
            self.start()
            atexit.register(self._disconnect_atexit)

    def start(self, timeout: int = 20):
        geth_dev = self._create_process()
        geth_dev.connect(timeout=timeout)
        if not self.web3.is_connected():
            geth_dev.disconnect()

            raise ConnectionError(f"Unable to connect to locally running {geth_dev.process_name}.")
        else:
            self.web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

        self._process = geth_dev

        # For subprocess-provider
        if self._process is not None and (process := self._process.proc):
            self.stderr_queue = JoinableQueue()
            self.stdout_queue = JoinableQueue()

            self.process = process

            # Start listening to output.
            spawn(self.produce_stdout_queue)
            spawn(self.produce_stderr_queue)
            spawn(self.consume_stdout_queue)
            spawn(self.consume_stderr_queue)

            self.network_manager.running_nodes.cache_provider(self)

    def _create_process(self) -> GethDevProcess:
        # NOTE: Using JSON mode to ensure types can be passed as CLI args.
        test_config = self.config_manager.get_config("test").model_dump(mode="json")

        # Allow configuring a custom executable besides your $PATH geth.
        if self.settings.executable is not None:
            test_config["executable"] = self.settings.executable

        test_config["ipc_path"] = self.ipc_path

        # Let the provider handle disconnecting the process.
        # This avoids multiple atexit handlers from being unnecessarily
        # registered that do some of the same thing.
        test_config["auto_disconnect"] = False

        # Include extra accounts to allocated funds to at genesis.
        extra_accounts = self.settings.ethereum.local.get("extra_funded_accounts", [])
        extra_accounts.extend(self.provider_settings.get("extra_funded_accounts", []))
        extra_accounts = list({a.lower() for a in extra_accounts})
        test_config["extra_funded_accounts"] = extra_accounts
        test_config["initial_balance"] = self.test_config.balance
        test_config["background"] = self.background
        uri = self.ws_uri or self.uri
        proc_data_dir = self.data_dir / f"{self.process_name}"
        return GethDevProcess.from_uri(
            uri,
            proc_data_dir,
            block_time=self.block_time,
            **test_config,
        )

    def disconnect(self):
        # Must disconnect process first.
        if self._process is not None:
            self._process.disconnect()
            self._process = None

            # Remove self from managed-processes list.
            self.network_manager.running_nodes.remove_provider(self)

        # Also unset the subprocess-provider reference.
        # NOTE: Type ignore is wrong; TODO: figure out why.
        self.process = None  # type: ignore[assignment]

        # Clear any snapshots.
        self.chain_manager._snapshots[self.chain_id] = []

        super().disconnect()

    def send_transaction(self, txn: "TransactionAPI") -> "ReceiptAPI":
        return self._send_transaction_with_retries(txn)

    def _send_transaction_with_retries(
        self, txn: "TransactionAPI", nonce_retries: int = 0, max_nonce_retries: int = 5
    ) -> "ReceiptAPI":
        try:
            return super().send_transaction(txn)
        except VirtualMachineError as err:
            if (
                txn.sender in self.account_manager.test_accounts
                and "exceeds block gas limit" in str(err)
            ):
                # Changed, possibly due to other transactions (x-dist?).
                # Retry using block gas limit.
                block_gas_limit = self.chain_manager.blocks.head.gas_limit
                if txn.gas_limit > block_gas_limit:
                    txn.gas_limit = block_gas_limit
                elif txn.gas_limit == block_gas_limit:
                    txn.gas_limit -= 1
                else:
                    # Raise whatever error it is. I am not sure how this is possible!
                    raise

                account = self.account_manager.test_accounts[txn.sender]
                signed_transaction = account.sign_transaction(txn)
                logger.debug("Gas-limit exceeds block gas limit. Retrying using block gas limit.")
                return super().send_transaction(signed_transaction)

            elif txn.sender in self.account_manager.test_accounts and re.match(
                r".*Nonce '\d*' is too low.*", str(err)
            ):
                retries = nonce_retries + 1
                if retries > max_nonce_retries:
                    raise  # This error.

                # Try again with a new nonce.
                account = self.account_manager.test_accounts[txn.sender]
                txn.nonce = account.nonce
                signed_transaction = account.sign_transaction(txn)
                logger.debug("Test transaction received bad nonce. Retrying using latest nonce.")
                return self._send_transaction_with_retries(
                    signed_transaction,
                    nonce_retries=retries,
                    max_nonce_retries=max_nonce_retries,
                )

            raise  # Whatever error it already is (Ape-ified from ape-ethereum.provider base).

    def snapshot(self) -> "SnapshotID":
        return self._get_latest_block().number or 0

    def restore(self, snapshot_id: "SnapshotID"):
        if isinstance(snapshot_id, int):
            block_number_int = snapshot_id
            block_number_hex_str = str(to_hex(snapshot_id))
        elif isinstance(snapshot_id, bytes):
            block_number_hex_str = add_0x_prefix(to_hex(snapshot_id))
            block_number_int = int(block_number_hex_str, 16)
        else:
            block_number_hex_str = to_hex(snapshot_id)
            block_number_int = int(snapshot_id, 16)

        current_block = self._get_latest_block().number or 0
        if block_number_int == current_block:
            # Head is already at this block.
            return
        elif block_number_int > current_block:
            logger.error("Unable to set head to future block.")
            return

        self.make_request("debug_setHead", [block_number_hex_str])

    @raises_not_implemented
    def set_timestamp(self, new_timestamp: int):
        pass

    @raises_not_implemented
    def mine(self, num_blocks: int = 1):
        pass

    def build_command(self) -> list[str]:
        return self._process.command if self._process else []

    def get_test_account(self, index: int) -> "TestAccountAPI":
        if self._process is None:
            # Not managing the process. Use default approach.
            test_container = self.account_manager.test_accounts.containers["test"]
            return test_container.generate_account(index)

        # perf: we avoid having to generate account keys twice by utilizing
        #   the accounts generated for geth's genesis.json.
        account = self._process._dev_accounts[index]
        return self.account_manager.init_test_account(index, account.address, account.private_key)


# NOTE: The default behavior of EthereumNodeBehavior assumes geth.
class Node(EthereumNodeProvider):
    pass



import re
from collections.abc import Iterable, Sequence
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from functools import cached_property
from itertools import islice
from types import UnionType
from typing import TYPE_CHECKING, Any, Union, get_args, get_origin

from cchecksum import to_checksum_address
from dateutil.parser import parse
from eth_pydantic_types import Address, HexBytes
from eth_typing.evm import ChecksumAddress
from eth_utils import is_0x_prefixed, is_checksum_address, is_hex, is_hex_address, to_int

from ape.api.address import BaseAddress
from ape.api.convert import ConverterAPI, ConvertibleAPI
from ape.api.transactions import TransactionAPI
from ape.exceptions import ConversionError
from ape.logging import logger
from ape.types.address import AddressType
from ape.utils.misc import log_instead_of_fail

from .base import BaseManager

if TYPE_CHECKING:
    from ethpm_types import ConstructorABI, EventABI, MethodABI


class HexConverter(ConverterAPI):
    """
    A converter that converts ``str`` to ``HexBytes``.
    NOTE: This utility converter ensures that all bytes args can accept hex too
    """

    def is_convertible(self, value: Any) -> bool:
        return (
            (isinstance(value, str) and is_hex(value))
            or isinstance(value, bytes)
            or isinstance(value, int)
        )

    def convert(self, value: str) -> bytes:
        """
        Convert the given value to ``HexBytes``.

        Args:
            value (str): The value to convert.

        Returns:
            bytes
        """

        return HexBytes(value)


class HexIntConverter(ConverterAPI):
    """
    Convert hex values to integers.

    **NOTE** If value is a ``str``, it must begin with "0x".
    """

    def is_convertible(self, value: Any) -> bool:
        # NOTE: Hex int conversion requires 0x prefix to know if it is base 16 or base 10.
        return (isinstance(value, str) and is_hex(value) and value.startswith("0x")) or isinstance(
            value, bytes
        )

    def convert(self, value: Any) -> int:
        return to_int(HexBytes(value))


class HexIterableConverter(ConverterAPI):
    """
    Convert list of hex values to single concatenated ``HexBytes`` value.
    """

    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, Iterable) and all(isinstance(v, bytes) or is_hex(v) for v in value)

    def convert(self, value: Any) -> bytes:
        return HexBytes(b"".join(HexBytes(v) for v in value))


class StringIntConverter(ConverterAPI):
    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, str) and not is_0x_prefixed(value) and value.isnumeric()

    def convert(self, value: str) -> int:
        return int(value)


class AccountIntConverter(ConverterAPI):
    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, BaseAddress)

    def convert(self, value: BaseAddress) -> int:
        return self.conversion_manager.convert(value.address, int)


class AddressAPIConverter(ConverterAPI):
    """
    A converter that converts an :class:`~ape.api.address.BaseAddress`
    to a :class`~ape.types.address.AddressType`.
    """

    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, BaseAddress)

    def convert(self, value: BaseAddress) -> AddressType:
        """
        Convert the given value to :class:`~ape.types.address.AddressType`.

        Args:
            value (str): The value to convert.

        Returns:
            :class:`~ape.types.address.AddressType`: An alias to
            `ChecksumAddress <https://eth-typing.readthedocs.io/en/latest/types.html#checksumaddress>`__.  # noqa: E501
        """

        return value.address


class HexAddressConverter(ConverterAPI):
    """
    A converter that converts a checksummed address ``str`` to a
    :class:`~ape.types.address.AddressType`.
    """

    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, str) and is_hex_address(value) and not is_checksum_address(value)

    def convert(self, value: str) -> AddressType:
        """
        Convert the given value to a :class:`~ape.types.address.AddressType`.

        Args:
            value (str): The address ``str`` to convert.

        Returns:
            :class:`~ape.types.address.AddressType`
        """

        return AddressType(to_checksum_address(value))


class BytesAddressConverter(ConverterAPI):
    """
    A converter that converts a raw bytes address to an :class:`~ape.types.address.AddressType`.
    """

    def is_convertible(self, value: Any) -> bool:
        return isinstance(value, bytes) and len(value) == 20

    def convert(self, value: bytes) -> AddressType:
        return to_checksum_address(value)


class IntAddressConverter(ConverterAPI):
    """
    A converter that converts an integer address to an :class:`~ape.types.address.AddressType`.
    """

    _cache: dict[int, AddressType | bool] = {}

    def is_convertible(self, value: Any) -> bool:
        if not isinstance(value, int):
            return False
        elif isinstance(self._cache.get(value), str):
            return True

        val = self._convert(value)
        self._cache[value] = val
        return isinstance(val, str)

    def convert(self, value: Any) -> AddressType:
        err_msg = f"Failed to convert '{value}' to 'AddressType'."
        if cached_val := self._cache.get(value):
            if not isinstance(cached_val, str):
                # Shouldn't get here in normal execution.
                raise ConversionError(err_msg)

            return cached_val

        # Shouldn't get here in normal execution.
        res = self._convert(value)
        self._cache[value] = res

        if not isinstance(res, str):
            raise ConversionError(err_msg)

        return res

    def _convert(self, value: int) -> AddressType | bool:
        try:
            val = Address.__eth_pydantic_validate__(value)
        except Exception:
            return False

        return AddressType(to_checksum_address(val))


class TimestampConverter(ConverterAPI):
    """
    Converts either a string, datetime object, or a timedelta object to a timestamp.
    No timezone required, but should be formatted to UTC.
    """

    def is_convertible(self, value: str | datetime | timedelta) -> bool:
        if not isinstance(value, (str, datetime, timedelta)):
            return False
        if isinstance(value, str):
            if " " not in value or len(value.split(" ")) > 2:
                return False
            else:
                try:
                    parse(value)
                except ValueError:
                    return False
        return True

    def convert(self, value: str | datetime | timedelta) -> int:
        if isinstance(value, str):
            return int(parse(value).replace(tzinfo=timezone.utc).timestamp())
        elif isinstance(value, datetime):
            return int(value.replace(tzinfo=timezone.utc).timestamp())
        elif isinstance(value, timedelta):
            return int((datetime.now(timezone.utc) + value).timestamp())
        else:
            raise ConversionError()


class StringDecimalConverter(ConverterAPI):
    """
    Convert string-formatted floating point values to `Decimal` type.
    """

    def is_convertible(self, value: Any) -> bool:
        # Matches only string-formatted floats with an optional sign character (+/-).
        # Leading and trailing zeros are required.
        # NOTE: `re.fullmatch` will only match the full string, so "1.0 ether" and "10.0 USDC"
        # will not be identified as convertible.
        return isinstance(value, str) and re.fullmatch(r"[+-]?\d+\.\d+", value) is not None

    def convert(self, value: str) -> Decimal:
        return Decimal(value)


class ConversionManager(BaseManager):
    """
    A singleton that manages all the converters.

    **NOTE**: typically, users will not interact with this class directly,
    but rather its ``convert()`` method, which is accessible from
    the root ``ape`` namespace.

    Usage example::

        from ape import convert

        amount = convert("1 gwei", int)
    """

    @log_instead_of_fail(default="<DependencyAPI>")
    def __repr__(self) -> str:
        return f"<{ConversionManager.__name__}>"

    @cached_property
    def _converters(self) -> dict[type, list[ConverterAPI]]:
        converters: dict[type, list[ConverterAPI]] = {
            AddressType: [
                AddressAPIConverter(),
                BytesAddressConverter(),
                HexAddressConverter(),
                IntAddressConverter(),
            ],
            bytes: [
                HexConverter(),
                HexIterableConverter(),
            ],
            int: [
                TimestampConverter(),
                HexIntConverter(),
                StringIntConverter(),
                AccountIntConverter(),
            ],
            Decimal: [StringDecimalConverter()],
            bool: [],
            str: [],
        }

        for plugin_name, (conversion_type, converter_class) in self.plugin_manager.converters:
            converter = converter_class()
            if conversion_type is ChecksumAddress:
                converters[AddressType].append(converter)
                return converters

            elif conversion_type not in converters:
                options = ", ".join([_get_type_name_from_type(t) for t in converters])
                raise ConversionError(f"Type '{conversion_type}' must be one of [{options}].")

            converters[conversion_type].append(converter)

        return converters

    def is_type(self, value: Any, to_type: type) -> bool:
        """
        Check if the value is the given type.
        If given an :class:`~ape.types.address.AddressType`, will also check
        that it is checksummed.

        Args:
            value (any): The value to check.
            to_type (type): The type to check against.

        Returns:
            bool: ``True`` when we consider the given value to be the given type.
        """
        return is_checksum_address(value) if to_type is AddressType else isinstance(value, to_type)

    def convert(self, value: Any, to_type: type | tuple | list) -> Any:
        """
        Convert the given value to the given type. This method accesses
        all :class:`~ape.api.convert.ConverterAPI` instances known to
        `ape`` and selects the appropriate one, so long that it exists.

        Raises:
            :class:`~ape.exceptions.ConversionError`: When there is not a registered
              converter for the given arguments.

        Args:
            value (any): The value to convert.
            to_type (to_type): The type to convert the value to.

        Returns:
            any: The same given value but with the new given type.
        """

        if isinstance(value, (list, tuple)) and isinstance(to_type, tuple):
            # We expected to convert a tuple type, so convert each item in the tuple.
            #
            # NOTE: Some callers provide *extra* items (e.g. struct-like tuples /
            # namedtuples with additional fields). Extra values should be ignored,
            # similar to how we ignore extra keys in dict-based struct inputs.
            #
            # However, providing *too few* items should still fail fast.
            trimmed_values = list(islice(value, len(to_type)))
            return [self.convert(v, t) for v, t in zip(trimmed_values, to_type, strict=True)]

        elif isinstance(value, (list, tuple)) and isinstance(to_type, list) and len(to_type) == 1:
            # We expected to convert an array type(dynamic or static),
            # so convert each item in the list.
            # NOTE: type for static and dynamic array is a single item
            #  list containing the type of the array.
            return [self.convert(v, to_type[0]) for v in value]

        elif isinstance(to_type, (list, tuple)):
            raise ConversionError(
                f"Value '{value}' must be a list or tuple when given multiple types."
            )

        elif to_type is ChecksumAddress:
            # Use our Annotated alias.
            return self.convert(value, AddressType)

        elif to_type not in self._converters:
            options = ", ".join([_get_type_name_from_type(t) for t in self._converters])
            raise ConversionError(f"Type '{to_type}' must be one of [{options}].")

        elif self.is_type(value, to_type) and not isinstance(value, (list, tuple)):
            # NOTE: Always process lists and tuples
            return value

        if isinstance(value, ConvertibleAPI) and value.is_convertible(to_type):
            return value.convert_to(to_type)

        return self._convert_using_converter_apis(value, to_type)

    def _convert_using_converter_apis(self, value: Any, to_type: type) -> Any:
        for converter in self._converters[to_type]:
            try:
                is_convertible = converter.is_convertible(value)
            except Exception as err:
                # If errors while checking if we can convert, log the error
                # and assume it's not convertible.
                converter_name = converter.__class__.__name__
                msg = f"Issue while checking `{converter_name}.is_convertible()`: {err}"
                logger.error(msg)
                continue

            if not is_convertible:
                continue

            try:
                return converter.convert(value)
            except Exception as err:
                try:
                    error_value = f" '{value}' (type={type(value)}) "
                except Exception:
                    error_value = " "

                message = f"Failed to convert{error_value}"
                if converter_type_name := getattr(type(converter), "__name__", None):
                    message = f"{message}using '{converter_type_name}'."

                raise ConversionError(message) from err

        raise ConversionError(f"No conversion registered to handle '{value}'.")

    def get_converters_by_type(self, converter_type: type) -> list[ConverterAPI]:
        """
        Get all the converters for the given type.

        Args:
            converter_type (type): The type to get converters for.

        Returns:
            list[ConverterAPI]: All registered converters for the given type.
        """
        return self._converters.get(converter_type, [])

    def get_converter(self, name: str) -> ConverterAPI:
        """
        Get a converter plugin by name.

        Args:
            name (str): The name of the converter.

        Returns:
            :class:`~ape.api.converters.ConverterAPI`: The converter.
        """
        for converter_ls in self._converters.values():
            for converter in converter_ls:
                if converter.name == name.lower():
                    return converter

        raise ConversionError("No converters with name '{name}'.'")

    def convert_method_args(
        self,
        abi: "MethodABI | ConstructorABI | EventABI",
        arguments: Sequence[Any],
    ):
        input_types = [i.canonical_type for i in abi.inputs]
        converted_arguments = []
        for ipt, argument in zip(input_types, arguments, strict=True):
            # Handle primitive-addresses separately since they may not occur
            # on the tuple-conversion if they are integers or bytes.
            if str(ipt) == "address":
                converted_value = self.convert(argument, AddressType)
                converted_arguments.append(converted_value)
            else:
                converted_arguments.append(argument)

        return converted_arguments

    def convert_method_kwargs(self, kwargs) -> dict:
        fields = TransactionAPI.__pydantic_fields__

        def get_real_type(type_):
            # Handle both old (Optional/Union) and new (|) syntax
            # Try old syntax first (Optional/Union) - uses _typevar_types
            all_types = getattr(type_, "_typevar_types", None)
            if all_types and isinstance(all_types, (list, tuple)) and len(all_types) > 0:
                # Old syntax found, use it
                pass
            else:
                # Try new syntax (| operator) - uses get_args/get_origin
                origin = get_origin(type_)
                if origin is not None and origin in (Union, UnionType):
                    args = get_args(type_)
                    if args:
                        all_types = list(args)
                    else:
                        # No args means it's not a union, return as-is
                        return type_
                else:
                    # Not a union type, return as-is
                    return type_

            # Filter out None/NoneType
            valid_types = [t for t in all_types if t is not type(None) and t is not None]
            if len(valid_types) == 1:
                # This is something like Optional[int] or int | None,
                # however, if the user provides a value,
                # we want to convert to the non-optional type.
                return valid_types[0]

            # Not sure if this is possible; the converter may fail.
            return valid_types

        annotations = {name: get_real_type(f.annotation) for name, f in fields.items()}
        kwargs_to_convert = {k: v for k, v in kwargs.items() if k == "sender" or k in fields}
        converted_fields = {}
        for field_name, value in kwargs_to_convert.items():
            type_ = AddressType if field_name == "sender" else annotations.get(field_name)
            if type_:
                try:
                    converted_value = self.convert(value, type_)
                except ConversionError:
                    # Ignore conversion errors and use the values as-is.
                    converted_value = value

            else:
                converted_value = value

            converted_fields[field_name] = converted_value

        return {**kwargs, **converted_fields}


def _get_type_name_from_type(var_type: type) -> str:
    if hasattr(var_type, "__args__") and var_type.__args__:
        # Is Annotated
        real_type = var_type.__args__[0]
        return _get_type_name_from_type(real_type)

    elif var_type is ChecksumAddress:
        return "AddressType"  # Use our alias

    elif hasattr(var_type, "__name__"):
        return var_type.__name__

    else:
        message = "Unable to deduce type name"
        try:
            str_value = f"{var_type}"
        except Exception:
            str_value = ""

        if str_value:
            message = f"{message} for {str_value}"

        raise TypeError(f"{message}.")



from eth_pydantic_types import HexBytes

from ape.contracts.base import ContractMethodHandler


def test_encode_input_list_for_struct(chain, mocker, owner, method_abi_with_struct_input):
    method = method_abi_with_struct_input
    handler = ContractMethodHandler(contract=mocker.MagicMock(), abis=[method])

    # NOTE: Purposely passing a list for the struct type here!
    #   this is to ensure we can use lists (instead of objects or dictionaries).
    params = [owner.address, b"foo", b"bar"]

    actual = handler.encode_input(owner.address, params, b"", [])
    acct = owner.address.replace("0x", "")
    expected = HexBytes(
        f"0x26e0a196000000000000000000000000{acct}000000000000000"
        f"0000000000000000000000000000000000000000000000080000000"
        f"0000000000000000000000000000000000000000000000000000000"
        f"1200000000000000000000000000000000000000000000000000000"
        f"000000000140000000000000000000000000{acct}666f6f0000000"
        f"0000000000000000000000000000000000000000000000000000000"
        f"0000000000000000000000000000000000000000000000000000000"
        f"0006000000000000000000000000000000000000000000000000000"
        f"0000000000000362617200000000000000000000000000000000000"
        f"0000000000000000000000000000000000000000000000000000000"
        f"0000000000000000000000000000000000000000000000000000000"
        f"00000000000000000000000000000000000000000"
    )
    assert actual == expected


def test_info(contract_instance):
    # The `setNumber()` ABI defined a NatSpec.
    abi = [x for x in contract_instance.contract_type.abi if getattr(x, "name", "") == "setNumber"][
        0
    ]
    handler = ContractMethodHandler(contract=contract_instance, abis=[abi])
    actual = handler.info
    expected_doc = (
        "Emits a `NumberChange` event with the previous number, "
        "the new number, and the previous block hash"
    )
    expected = f"""
setNumber(uint256 num)
  @custom:emits {expected_doc}
  @custom:modifies Sets the `myNumber` state variable
  @custom:require num Must not be equal to 5
  @details Only the owner can call this function. The new number cannot be 5.
  @param num uint256 The new number to be set
""".strip()

    # Doing it like this because sometimes Vyper compiler puts them out
    # of order it seems...
    for line in expected.splitlines():
        assert line in actual


def test_info_no_natspec(project, vyper_contract_instance):
    # The `myNumber()` ABI does not have a natspec.
    abi = [
        x for x in project.VyperContract.contract_type.abi if getattr(x, "name", "") == "myNumber"
    ][0]
    handler = ContractMethodHandler(contract=vyper_contract_instance, abis=[abi])
    actual = handler.info
    assert actual == ""



from tests.conftest import geth_process_test


@geth_process_test
def test_get_code(mocker, chain, geth_contract, mock_sepolia):
    # NOTE: Using mock_sepolia because code doesn't get cached in local networks.
    actual = chain.get_code(geth_contract.address)
    expected = chain.provider.get_code(geth_contract.address)
    assert actual == expected

    # Ensure uses cache (via not using provider).
    provider_spy = mocker.spy(chain.provider.web3.eth, "get_code")
    _ = chain.get_code(geth_contract.address)
    assert provider_spy.call_count == 0

    # block_id test, cache should interfere.
    actual_2 = chain.get_code(geth_contract.address, block_id=0)
    assert not actual_2  # Doesn't exist at block 0.



import itertools
from datetime import datetime
from html.parser import HTMLParser
from pathlib import Path
from typing import TYPE_CHECKING, Any
from xml.dom.minidom import getDOMImplementation
from xml.etree.ElementTree import Element, SubElement, tostring

import requests
from ethpm_types.source import SourceLocation
from pydantic import NonNegativeInt, field_validator

from ape.logging import logger
from ape.utils.basemodel import BaseModel
from ape.utils.misc import get_current_timestamp_ms
from ape.version import version as ape_version

if TYPE_CHECKING:
    from ethpm_types.source import ContractSource


_APE_DOCS_URL = "https://docs.apeworx.io/ape/stable/index.html"
_DTD_URL = "https://raw.githubusercontent.com/cobertura/web/master/htdocs/xml/coverage-04.dtd"
_CSS = """
@import url("https://fonts.cdnfonts.com/css/barlow");

html {
  font-weight: 400;
  font-family: "Barlow", sans-serif;
  min-height: 100%;
  background: #8DCAEF;
  font-size: 17px;
}

body {
  font-weight: 400;
  font-family: "Barlow", sans-serif;
  min-height: 100%;
  font-size: 17px;
}

p {
  font-weight: 400;
  font-family: "Barlow", sans-serif;
  font-size: 18px;
  text-align: center;
  padding-bottom: 10px;
}

h1 {
  font-size: 60px;
  line-height: 54px;
  font-weight: 600;
  font-family: "Barlow", sans-serif;
  text-transform: uppercase;
  letter-spacing: -0.05em;
  color: #29B285;
  padding: 0 !important;
  text-align: center;
}

h2 {
  font-size: 30px;
  line-height: 27px;
  font-weight: 600;
  font-family: "Barlow", sans-serif;
  text-transform: uppercase;
  letter-spacing: -0.05em;
  color: #29B285;
  padding: 0 !important;
  display: flex;
  justify-content: space-between;
}

.left-aligned {
  text-align: left;
}

.right-aligned {
  text-align: right;
}

.table-center {
  width: 60%;
  margin: 0 auto;
  padding-left: 10%;
  padding-right: 10%;
}

table {
  width: 80%;
  margin: 0 auto;
  padding-left: 10%;
  padding-right: 10%;
  padding-bottom: 64px;
}

th,
td {
  padding: 8px;
}

th.column1,
td.column1 {
  text-align: left;
}

th.column2,
td.column2,
th.column3,
td.column3,
th.column4,
td.column4,
th.column5,
td.column5 {
  text-align: right;
}
""".lstrip()


class CoverageStatement(BaseModel):
    """
    An item that can get hit during coverage. Examples of coverage items are
    line segments, which are generally calculated from groupings of AST nodes
    occupying the same location, that can be tracked by their PC values. During
    a transaction's trace, we find these values and we find the corresponding
    coverage item and increment the hit count. Another example of a coverage item
    is a compiler-builtin check, also marked by a PC value. If we encounter such
    PC in a trace, its hit count is incremented. Builtin compiler checks may or
    may not correspond to an actual location in the source code, depending on the
    type of check.
    """

    location: SourceLocation | None = None
    """
    The location of the item (line, column, endline, endcolumn).
    If multiple PCs share an exact location, it is only tracked as one.
    """

    pcs: set[int]
    """
    The PCs for this node.
    """

    hit_count: NonNegativeInt = 0
    """
    The times this node was hit.
    """

    tag: str | None = None
    """
    An additional tag to mark this statement with.
    This is useful if the location field is empty.
    """


class FunctionCoverage(BaseModel):
    """
    The individual coverage of a function defined in a smart contact.
    """

    name: str
    """
    The display name of the function.
    """

    full_name: str
    """
    The unique name of the function.
    """

    statements: list[CoverageStatement] = []
    """
    For statement coverage, these are the individual items.
    See :class:`~ape.types.coverage.CoverageStatement` for more details.
    """

    hit_count: NonNegativeInt = 0
    """
    The times this function was called.
    **NOTE**: This is needed as a separate data point since not all methods may have
    statements (such as auto-getters).
    """

    @property
    def lines_covered(self) -> NonNegativeInt:
        """
        The number of lines with a hit counter greater than zero in this method.
        """
        return len([x for x in self.statements if x.hit_count > 0])

    @property
    def lines_valid(self) -> NonNegativeInt:
        """
        All lines valid for coverage in this method.
        """
        return len(self.statements)

    @property
    def miss_count(self) -> NonNegativeInt:
        """
        The number of lines missed.
        """
        return self.lines_valid - self.lines_covered

    @property
    def line_rate(self) -> float:
        """
        The number of lines hit divided by number of lines.
        """
        if not self.statements:
            # If there are no statements, rely on fn hit count only.
            return 1.0 if self.hit_count > 0 else 0.0

        # This function has hittable statements.
        return self.lines_covered / self.lines_valid if self.lines_valid > 0 else 0

    def model_dump(self, *args, **kwargs) -> dict:
        attribs = super().model_dump(*args, **kwargs)

        # Add coverage stats.
        attribs["lines_covered"] = self.lines_covered
        attribs["lines_valid"] = self.lines_valid
        attribs["line_rate"] = self.line_rate

        return attribs

    def profile_statement(
        self, pc: int, location: SourceLocation | None = None, tag: str | None = None
    ):
        """
        Initialize a statement in the coverage profile with a hit count starting at zero.
        This statement is ready to accumulate hits as tests execute.

        Args:
            pc (int): The program counter of the statement.
            location (Optional[ethpm_types.source.SourceStatement]): The location of the statement,
              if it exists.
            tag (Optional[str]): Optionally provide more information about the statements being hit.
              This is useful for builtin statements that may be missing context otherwise.
        """

        for statement in self.statements:
            if not location or (
                location and statement.location and statement.location[0] != location[0]
            ):
                # Starts on a different line.
                continue

            # Already tracking this location.
            statement.pcs.add(pc)

            if not statement.tag:
                statement.tag = tag

            return

        if location:
            # Adding a source-statement for the first time.
            coverage_statement = CoverageStatement(location=location, pcs={pc}, tag=tag)

        else:
            # Adding a virtual statement.
            coverage_statement = CoverageStatement(pcs={pc}, tag=tag)

        if coverage_statement is not None:
            self.statements.append(coverage_statement)


class ContractCoverage(BaseModel):
    """
    An individual contract's coverage.
    """

    name: str
    """
    The name of the contract.
    """

    functions: list[FunctionCoverage] = []
    """
    The coverage of each function individually.
    """

    @property
    def statements(self) -> list[CoverageStatement]:
        """
        All valid coverage lines from every function in this contract.
        """
        return list(itertools.chain.from_iterable(f.statements for f in self.functions))

    @property
    def lines_covered(self) -> NonNegativeInt:
        """
        All lines that have a hit count greater than zero.
        """

        return sum(funcs.lines_covered for funcs in self.functions)

    @property
    def lines_valid(self) -> NonNegativeInt:
        """
        The number of lines valid for coverage.
        """
        return len(self.statements)

    @property
    def miss_count(self) -> NonNegativeInt:
        """
        The number of lines missed.
        """
        return self.lines_valid - self.lines_covered

    @property
    def line_rate(self) -> float:
        """
        The number of lines hit divided by number of lines.
        """
        return self.lines_covered / self.lines_valid

    @property
    def function_hits(self) -> NonNegativeInt:
        """
        The number of functions with a hit counter greater than zero.
        """
        return len([fn for fn in self.functions if fn.hit_count > 0])

    @property
    def function_rate(self) -> float:
        """
        The rate of functions hit versus total functions.
        """
        return self.function_hits / len(self.functions)

    def __getitem__(self, function_name: str) -> FunctionCoverage:
        func = self.get_function(function_name)
        if func:
            return func

        raise KeyError(f"Function '{function_name}' not found.")

    def model_dump(self, *args, **kwargs) -> dict:
        attribs = super().model_dump(*args, **kwargs)

        # Add coverage stats.
        attribs["lines_covered"] = self.lines_covered
        attribs["lines_valid"] = self.lines_valid
        attribs["line_rate"] = self.line_rate

        return attribs

    def include(self, name: str, full_name: str) -> FunctionCoverage:
        # Make sure function is included in coverage.
        func_cov = self.get_function(full_name)
        if func_cov:
            return func_cov

        func_cov = FunctionCoverage(name=name, full_name=full_name)
        self.functions.append(func_cov)
        return func_cov

    def get_function(self, full_name: str) -> FunctionCoverage | None:
        for func in self.functions:
            if func.full_name == full_name:
                return func

        return None


class ContractSourceCoverage(BaseModel):
    """
    An individual source file with coverage collected.
    """

    source_id: str
    """
    The ID of the source covered.
    """

    contracts: list[ContractCoverage] = []
    """
    Coverage for each contract in the source file.
    """

    @property
    def statements(self) -> list[CoverageStatement]:
        """
        All valid coverage lines from every function in every contract in this source.
        """

        return list(itertools.chain.from_iterable(c.statements for c in self.contracts))

    @property
    def lines_covered(self) -> NonNegativeInt:
        """
        All lines with a hit count greater than zero from every function
        in every contract in this source.
        """
        return sum(c.lines_covered for c in self.contracts)

    @property
    def lines_valid(self) -> NonNegativeInt:
        """
        The number of lines valid for coverage.
        """
        return len(self.statements)

    @property
    def miss_count(self) -> NonNegativeInt:
        """
        The number of lines missed.
        """
        return self.lines_valid - self.lines_covered

    @property
    def line_rate(self) -> float:
        """
        The number of lines hit divided by number of lines.
        """
        return self.lines_covered / self.lines_valid if self.lines_valid > 0 else 0

    @property
    def total_functions(self) -> NonNegativeInt:
        """
        The total number of functions in this source.
        """
        return sum(len(c.functions) for c in self.contracts)

    @property
    def function_hits(self) -> NonNegativeInt:
        """
        The number of functions with a hit counter greater than zero.
        """
        return sum(c.function_hits for c in self.contracts)

    @property
    def function_rate(self) -> float:
        """
        The rate of functions hit versus total functions.
        """
        return self.function_hits / self.total_functions if self.total_functions > 0 else 0

    def model_dump(self, *args, **kwargs) -> dict:
        attribs = super().model_dump(*args, **kwargs)

        # Add coverage stats.
        attribs["lines_covered"] = self.lines_covered
        attribs["lines_valid"] = self.lines_valid
        attribs["line_rate"] = self.line_rate

        return attribs

    def include(self, contract_name: str) -> ContractCoverage:
        """
        Ensure a contract is included in the report.
        """

        for contract in self.contracts:
            if contract.name == contract_name:
                return contract

        # Include the contract.
        contract_cov = ContractCoverage(name=contract_name)
        self.contracts.append(contract_cov)
        return contract_cov


class CoverageProject(BaseModel):
    """
    A project with coverage collected.
    """

    name: str
    """
    The name of the project being covered.
    """

    sources: list[ContractSourceCoverage] = []
    """
    Coverage for each source in the project.
    """

    @property
    def statements(self) -> list[CoverageStatement]:
        """
        All valid coverage lines from every function in every contract in every source
        in this project.
        """

        return list(itertools.chain.from_iterable(s.statements for s in self.sources))

    @property
    def lines_covered(self) -> NonNegativeInt:
        """
        The number of lines with a hit count greater than zero from every function
        in every contract in every source in this this project.
        """
        return sum(s.lines_covered for s in self.sources)

    @property
    def lines_valid(self) -> NonNegativeInt:
        """
        The number of lines valid for coverage.
        """
        return len(self.statements)

    @property
    def miss_count(self) -> NonNegativeInt:
        """
        The number of lines missed.
        """
        return self.lines_valid - self.lines_covered

    @property
    def line_rate(self) -> float:
        """
        The number of lines hit divided by number of lines.
        """
        return self.lines_covered / self.lines_valid if self.lines_valid > 0 else 0

    @property
    def total_functions(self) -> NonNegativeInt:
        """
        The total number of functions in this source.
        """
        return sum(x.total_functions for x in self.sources)

    @property
    def function_hits(self) -> NonNegativeInt:
        """
        The number of functions with a hit counter greater than zero.
        """
        return sum(x.function_hits for x in self.sources)

    @property
    def function_rate(self) -> float:
        """
        The rate of functions hit versus total functions.
        """
        return self.function_hits / self.total_functions if self.total_functions > 0 else 0

    def model_dump(self, *args, **kwargs) -> dict:
        attribs = super().model_dump(*args, **kwargs)

        # Add coverage stats.
        attribs["lines_covered"] = self.lines_covered
        attribs["lines_valid"] = self.lines_valid
        attribs["line_rate"] = self.line_rate

        return attribs

    def include(self, contract_source: "ContractSource") -> ContractSourceCoverage:
        for src in self.sources:
            if src.source_id == contract_source.source_id:
                return src

        source_cov = ContractSourceCoverage(source_id=contract_source.source_id)
        self.sources.append(source_cov)
        return source_cov


class CoverageReport(BaseModel):
    """
    Coverage report schema inspired from coverage.py.
    """

    source_folders: list[Path]
    """
    All source folders to use. This is needed for codecov.
    """

    timestamp: int
    """
    The timestamp the report was generated, in milliseconds.
    """

    projects: list[CoverageProject] = []
    """
    Each project with individual coverage tracked.
    """

    @field_validator("timestamp", mode="before")
    @classmethod
    def validate_timestamp(cls, value):
        # Default to current UTC timestamp (ms).
        return value or get_current_timestamp_ms()

    @property
    def sources(self) -> list[str]:
        """
        Every source ID in the report.
        """
        return [s.source_id for p in self.projects for s in p.sources]

    @property
    def statements(self) -> list[CoverageStatement]:
        """
        All valid coverage lines from every function in every contract in every source
        from every project in this report.
        """
        return list(itertools.chain.from_iterable(p.statements for p in self.projects))

    @property
    def lines_covered(self) -> NonNegativeInt:
        """
        All lines with a hit count greater than zero from every function
        in every contract in every source in every project in this report.
        """
        return sum(p.lines_covered for p in self.projects)

    @property
    def lines_valid(self) -> NonNegativeInt:
        """
        The number of lines valid for coverage.
        """
        return len(self.statements)

    @property
    def miss_count(self) -> NonNegativeInt:
        """
        The number of lines missed.
        """
        return self.lines_valid - self.lines_covered

    @property
    def line_rate(self) -> float:
        """
        The number of lines hit divided by number of lines.
        """
        return self.lines_covered / self.lines_valid if self.lines_valid > 0 else 0

    @property
    def total_functions(self) -> NonNegativeInt:
        """
        The total number of functions in this source.
        """
        return sum(x.total_functions for x in self.projects)

    @property
    def function_hits(self) -> NonNegativeInt:
        """
        The number of functions with a hit counter greater than zero.
        """
        return sum(x.function_hits for x in self.projects)

    @property
    def function_rate(self) -> float:
        """
        The rate of functions hit versus total functions.
        """
        return self.function_hits / self.total_functions if self.total_functions > 0 else 0

    def get_xml(self) -> str:
        """
        The coverage XML report as a string. The XML coverage data schema is
        meant to be compatible with codecov.io. Thus, some of coverage is modified
        slightly, and some of the naming conventions (based on 90s Java) won't be
        super relevant to smart-contract projects.
        """
        # See _DTD_URL to learn more about the schema.
        xml_out = self._get_xml()
        return xml_out.toprettyxml(indent="  ")

    def _get_xml(self):
        # NOTE: Some of this implementation is borrowed from coverage.py.
        impl = getDOMImplementation()
        if not impl:
            # Only for mypy.
            raise ValueError("Failed to get XML DOM.")

        xml_out = impl.createDocument(None, "coverage", None)
        if not (xcoverage := xml_out.documentElement):
            return xml_out

        # Unable to use too exotic of a version.
        xversion = ape_version.split(".dev")[0].strip()
        xcoverage.setAttribute("version", xversion)

        # Add top-level statement stats.
        xcoverage.setAttribute("timestamp", f"{self.timestamp}")
        xcoverage.setAttribute("lines-valid", f"{self.lines_valid}")
        xcoverage.setAttribute("lines-covered", f"{self.lines_covered}")
        xcoverage.setAttribute("line-rate", f"{round(self.line_rate, 4)}")

        # NOTE: Branch fields are required in the schema.
        # TODO: Replace with actual branch coverage when exists.
        xcoverage.setAttribute("branches-covered", "0")
        xcoverage.setAttribute("branches-valid", "0")
        xcoverage.setAttribute("branch-rate", "0")

        # I don't know what this, but it is also required.
        xcoverage.setAttribute("complexity", "0")

        # Add comments.
        xcoverage.appendChild(
            xml_out.createComment(f" Generated by Ape Framework: {_APE_DOCS_URL}")
        )
        xcoverage.appendChild(xml_out.createComment(f" Based on  {_DTD_URL} "))

        # In the XML schema, sources refer to root directories containing source code.
        # In our case, that would typically be the "contracts" folder of the project.
        # NOTE: This is critical and necessary for codecov to map sources correctly.
        xsources = xml_out.createElement("sources")
        for source_path in self.source_folders:
            xsource = xml_out.createElement("source")
            xtxt = xml_out.createTextNode(source_path.as_posix())
            xsource.appendChild(xtxt)
            xsources.appendChild(xsource)

        xcoverage.appendChild(xsources)

        # projects = packages.
        xpackages = xml_out.createElement("packages")
        for project in self.projects:
            xpackage = xml_out.createElement("package")

            # NOTE: For now, always use "." as the package name.
            # TODO: Experiment with using `self.project.name` as package name.
            # If it is "__local__", definitely use "." instead here.
            xpackage.setAttribute("name", ".")

            # Add package-level stats.
            xpackage.setAttribute("line-rate", f"{round(project.line_rate, 4)}")
            xpackage.setAttribute("branch-rate", "0")  # TODO
            xpackage.setAttribute("complexity", "0")

            # The `classes` field refers to `contracts` in our case.
            xclasses = xml_out.createElement("classes")
            for src in project.sources:
                for contract in src.contracts:
                    xclass = xml_out.createElement("class")
                    xclass.setAttribute("name", src.source_id)
                    xclass.setAttribute("line-rate", f"{round(contract.line_rate, 4)}")
                    xclass.setAttribute("branch-rate", "0")  # TODO
                    xclass.setAttribute("complexity", "0")

                    # NOTE: I am not sure what this does or why it is needed.
                    # Also, I am not sure why we don't map statements to the methods.
                    # because we totally could do that. Nonetheless, we have to follow
                    # the schema.
                    xml_out.createElement("methods")

                    # Use name unless the same function found twice, then use full name.
                    fn_map: dict[str, FunctionCoverage] = {}
                    fn_singles_used = []

                    # For the XML report, we split all statements to be only 1 line long.
                    # Each class (contract) can only identify the statement (line number) once.
                    lines_to_add: dict[int, int] = {}
                    xlines = xml_out.createElement("lines")

                    for function in contract.functions:
                        fn_singles_used.append(function.name)
                        if (
                            function.name in fn_map
                            and function.full_name != fn_map[function.name].full_name
                        ):
                            # Another method with the same name already in map.
                            # Use full name for both.
                            existing_fn = fn_map[function.name]
                            fn_map[existing_fn.full_name] = existing_fn
                            del fn_map[function.name]
                            fn_map[function.full_name] = function
                        elif function.name in fn_singles_used:
                            # Because this name has already been found once,
                            # we can assume we are using full names for these.
                            fn_map[function.full_name] = function
                        else:
                            # Is first time coming across this name.
                            fn_map[function.name] = function

                    for fn_name, function in fn_map.items():
                        if not function.statements or not any(
                            s.location for s in function.statements
                        ):
                            # Functions without source-locatable statements are not
                            # permitted in the XML report. This mean auto-getter hits
                            # won't be included as well as any builtin lines. Use other
                            # reportsF to find that level of information.
                            continue

                        for statement in function.statements:
                            if not statement.location:
                                # Statements without line numbers are excluded from this report.
                                # That level of granularity is present in other reports however.
                                # The XML report is strict so it can merge with others.
                                continue

                            for lineno in range(statement.location[0], statement.location[2] + 1):
                                if lineno in lines_to_add:
                                    lines_to_add[lineno] += statement.hit_count
                                else:
                                    lines_to_add[lineno] = statement.hit_count

                    # NOTE: Line numbers must be sorted in the XML!
                    sorted_nos = sorted(list(lines_to_add.keys()))
                    for no in sorted_nos:
                        hits = lines_to_add[no]
                        xline = xml_out.createElement("line")
                        xline.setAttribute("number", f"{no}")
                        xline.setAttribute("hits", f"{hits}")

                        xlines.appendChild(xline)
                        xclass.appendChild(xlines)
                    xclass.appendChild(xlines)
                    xclasses.appendChild(xclass)
            xpackage.appendChild(xclasses)
            xpackages.appendChild(xpackage)
        xcoverage.appendChild(xpackages)
        return xml_out

    def write_xml(self, path: Path):
        if not (xml := self.get_xml()):
            return

        elif path.is_dir():
            path = path / "coverage.xml"

        path.unlink(missing_ok=True)
        path.write_text(xml, encoding="utf8")

    def write_html(self, path: Path, verbose: bool = False):
        if not (html := self.get_html(verbose=verbose)):
            return

        elif path.is_dir():
            # Use as base path if given.
            html_path = path / "htmlcov"
            html_path.mkdir(exist_ok=True)

        elif not path.exists() and path.parent.is_dir():
            # Write to path if given a new one.
            html_path = path

        else:
            raise ValueError("Invalid path argument to `write_html()`.")

        # Create new index.html.
        index = html_path / "index.html"
        index.unlink(missing_ok=True)
        index.write_text(html, encoding="utf8")

        favicon = html_path / "favicon.ico"
        if not favicon.is_file():
            # Use favicon that already ships with Ape's docs.
            root = Path(__file__).parent
            docs_folder = root / "docs"
            while "ape" in root.as_posix() and not docs_folder.is_dir():
                root = root.parent
                docs_folder = root / "docs"

            docs_favicon = docs_folder / "favicon.ico"
            if docs_folder.is_dir() and docs_favicon.is_file():
                favicon.write_bytes(docs_favicon.read_bytes())
            else:
                # Try downloading from the internet. This may happen if running
                # ape in an isolated file system or a temporary directory,
                # such as CI/CD tests for Ape.
                try:
                    url = "https://github.com/ApeWorX/ape/blob/main/docs/favicon.ico"
                    response = requests.get(url)
                    response.raise_for_status()  # Check for any errors during the request
                    favicon.write_bytes(response.content)
                except Exception as err:
                    # Don't let this stop us from generating the report.
                    logger.debug(f"Failed finding favicon for coverage HTML. {err}")

            css = html_path / "styles.css"
            css.unlink(missing_ok=True)
            css.write_text(_CSS, encoding="utf8")

    def get_html(self, verbose: bool = False) -> str:
        """
        The coverage HTML report as a string.
        """
        html = self._get_html(verbose=verbose)
        html_str = tostring(html, encoding="utf8", method="html").decode()
        return _HTMLPrettfier().prettify(html_str)

    def _get_html(self, verbose: bool = False) -> Any:
        html = Element("html")
        head = SubElement(html, "head")
        meta = SubElement(head, "meta")
        meta.set("http-equiv", "Content-Type")
        meta.set("content", "text/html; charset=utf-8")
        title = SubElement(head, "title")
        title.text = "Contract Coverage Report"
        favicon = SubElement(head, "link")
        favicon.set("rel", "icon")
        favicon.set("sizes", "32x32")
        favicon.set("href", "favicon.ico")
        css = SubElement(head, "link")
        css.set("rel", "stylesheet")
        css.set("href", "styles.css")
        SubElement(html, "body")
        self._html_header_sub_element(html)
        self._html_main_sub_element(html, verbose=verbose)
        return html

    def _html_header_sub_element(self, html: Any) -> Any:
        header = SubElement(html, "header")
        div = SubElement(header, "div")
        h1 = SubElement(div, "h1")
        h1.text = "Coverage report"

        if len(self.projects) == 1:
            # If only one project, include information here instead of below.
            h1.text += f" for {self.projects[0].name}"

        paragraph = SubElement(div, "p")
        datetime_obj = datetime.fromtimestamp(self.timestamp / 1000)
        datetime_string = datetime_obj.strftime("%Y-%m-%d %H:%M:%S")
        paragraph.text = f"Generated by Ape Framework v{ape_version}, {datetime_string}"

    def _html_main_sub_element(self, html: Any, verbose: bool = False):
        main = SubElement(html, "main")
        show_project_header = len(self.projects) > 1
        if verbose:
            self._html_main_verbose_sub_element(main, show_project_header=show_project_header)
        else:
            self._html_main_non_verbose_sub_element(main, show_project_header=show_project_header)

    def _html_main_non_verbose_sub_element(self, main: Any, show_project_header: bool = True):
        columns = ("Source", "Statements", "Missing", "Statement Coverage", "Function Coverage")

        for project in self.projects:
            if show_project_header:
                # If only 1 project, it is shown at the top.
                title = SubElement(main, "h2")
                title.text = project.name

            table = SubElement(main, "table", {})
            thread = SubElement(table, "thread")
            thread_tr = SubElement(thread, "tr")

            for idx, column in enumerate(columns):
                th = SubElement(thread_tr, "th", {}, **{"class": f"column{idx + 1}"})
                th.text = column

            tbody = SubElement(table, "tbody")
            for src in project.sources:
                tbody_tr = SubElement(tbody, "tr")
                source_td = SubElement(tbody_tr, "td", {}, **{"class": "column1"})
                source_td.text = src.source_id
                self._set_common_td(tbody_tr, src)
                fn_cov_td = SubElement(tbody_tr, "td", {}, **{"class": "column5"})
                fn_cov_td.text = f"{round(src.function_rate * 100, 2)}%"

    def _html_main_verbose_sub_element(self, main: Any, show_project_header: bool = True):
        columns = (
            "Source",
            "Statements",
            "Missing",
            "Statement Coverage",
        )

        for project in self.projects:
            if show_project_header:
                # If only 1 project, it is shown at the top.
                title = SubElement(main, "h2")
                title.text = str(project.name)
                src_type = "h3"
            else:
                src_type = "h2"

            for src in project.sources:
                table_header_h = SubElement(main, src_type, {}, **{"class": "table-center"})
                stmt_cov = f"{round(src.line_rate * 100, 2)}%"
                fn_cov = f"{round(src.function_rate * 100, 2)}%"
                left_span = SubElement(table_header_h, "span", {}, **{"class": "left-aligned"})
                left_span.text = src.source_id
                right_span = SubElement(table_header_h, "span", {}, **{"class": "right-aligned"})
                right_span.text = f"stmt={stmt_cov} function={fn_cov}"
                table = SubElement(main, "table")
                thread = SubElement(table, "thread")
                thread_tr = SubElement(thread, "tr")

                for idx, column in enumerate(columns):
                    th = SubElement(thread_tr, "th", {}, **{"class": f"column{idx + 1}"})
                    th.text = column

                tbody = SubElement(table, "tbody")

                for contract in src.contracts:
                    for function in contract.functions:
                        tbody_tr = SubElement(tbody, "tr")
                        function_td = SubElement(tbody_tr, "td", {}, **{"class": "column1"})

                        # NOTE: Use the full name if the short name is repeated.
                        function_td.text = (
                            function.full_name
                            if len([fn for fn in contract.functions if fn.name == function.name])
                            > 1
                            else function.name
                        )

                        self._set_common_td(tbody_tr, function)

    def _set_common_td(self, tbody_tr: Any, src_or_fn: Any):
        stmts_td = SubElement(tbody_tr, "td", {}, **{"class": "column2"})
        stmts_td.text = f"{src_or_fn.lines_valid}"
        missing_td = SubElement(tbody_tr, "td", {}, **{"class": "column3"})
        missing_td.text = f"{src_or_fn.miss_count}"
        stmt_cov_td = SubElement(tbody_tr, "td", {}, **{"class": "column4"})
        stmt_cov_td.text = f"{round(src_or_fn.line_rate * 100, 2)}%"

    def model_dump(self, *args, **kwargs) -> dict:
        attribs = super().model_dump(*args, **kwargs)

        # Add coverage stats.
        attribs["lines_covered"] = self.lines_covered
        attribs["lines_valid"] = self.lines_valid
        attribs["line_rate"] = self.line_rate

        return attribs

    def get_source_coverage(self, source_id: str) -> ContractSourceCoverage | None:
        for project in self.projects:
            for src in project.sources:
                if src.source_id == source_id:
                    return src

        return None


class _HTMLPrettfier(HTMLParser):
    def __init__(self):
        super().__init__()
        self.no_indent_tags = (
            "html",
            "meta",
        )
        self.no_newline_tags = ("title", "h1", "h2", "th", "td")

        # State variables - get modified during prettification.
        self.indent = 0
        self.prettified_html = "<!DOCTYPE html>\n"

    def prettify(self, html_str: str) -> str:
        """
        This is a custom method not part of the HTMLParser spec
        that ingests a coverage HTML str, handles the formatting, returns it,
        and resets this formatter's instance, so that the operation
        is more functional.
        """
        self.feed(html_str)
        result = self.prettified_html
        self.reset()
        self.indent = 0
        self.prettified_html = "<!DOCTYPE html>\n"
        return result

    def handle_starttag(self, tag, attrs):
        self.prettified_html += " " * self.indent + "<" + tag
        for attr in attrs:
            self.prettified_html += f' {attr[0]}="{attr[1]}"'

        self.prettified_html += ">"
        if tag not in self.no_newline_tags:
            self.prettified_html += "\n"

        if tag not in self.no_indent_tags:
            self.indent += 2

    def handle_endtag(self, tag):
        self.indent = max(0, self.indent - 2)
        end_tag = f"</{tag}>\n"
        indented = self.prettified_html.endswith("\n")
        content = " " * self.indent + f"</{tag}>\n" if indented else end_tag
        self.prettified_html += content

    def handle_data(self, data):
        data_str = data.strip()
        if not data_str:
            return

        indented = self.prettified_html.endswith("\n")
        content = " " * self.indent + data + "\n" if indented else data
        self.prettified_html += content

    def handle_comment(self, data):
        self.prettified_html += " " * self.indent + f"<!--{data}-->\n"



from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.misc import raises_not_implemented


class BaseManager(ManagerAccessMixin):
    """
    Base manager that allows us to add other IPython integration features
    """

    @raises_not_implemented
    def _repr_mimebundle_(self, include=None, exclude=None):
        # This works better than AttributeError for Ape.
        pass

    @raises_not_implemented
    def _ipython_display_(self, include=None, exclude=None):
        # This works better than AttributeError for Ape.
        pass



from collections.abc import Iterator
from functools import singledispatchmethod

from ape.api.query import ContractCreation, ContractCreationQuery, QueryAPI, QueryType
from ape.exceptions import QueryEngineError
from ape.types.address import AddressType
from ape_ethereum.provider import EthereumNodeProvider


class OtterscanQueryEngine(QueryAPI):
    @singledispatchmethod
    def estimate_query(self, query: QueryType) -> int | None:  # type: ignore[override]
        return None

    @singledispatchmethod
    def perform_query(self, query: QueryType) -> Iterator:  # type: ignore[override]
        raise QueryEngineError(
            f"{self.__class__.__name__} cannot handle {query.__class__.__name__} queries."
        )

    @estimate_query.register
    def estimate_contract_creation_query(self, query: ContractCreationQuery) -> int | None:
        if getattr(self.provider, "_ots_api_level", None) is not None:
            return 250
        return None

    @perform_query.register
    def get_contract_creation_receipt(
        self, query: ContractCreationQuery
    ) -> Iterator[ContractCreation]:
        if self.network_manager.active_provider and isinstance(self.provider, EthereumNodeProvider):
            ots = self.provider.make_request("ots_getContractCreator", [query.contract])
            if ots is None:
                return None
            creator = self.conversion_manager.convert(ots["creator"], AddressType)
            receipt = self.provider.get_receipt(ots["hash"])
            yield ContractCreation(
                txn_hash=ots["hash"],
                block=receipt.block_number,
                deployer=receipt.sender,
                factory=creator if creator != receipt.sender else None,
            )



from collections.abc import Generator, Iterable, Iterator
from functools import cached_property
from importlib import import_module
from itertools import chain
from typing import Any

from ape.exceptions import ApeAttributeError
from ape.logging import logger
from ape.plugins._utils import CORE_PLUGINS, clean_plugin_name, get_plugin_dists
from ape.plugins.pluggy_patch import plugin_manager as pluggy_manager
from ape.utils.basemodel import _assert_not_ipython_check, only_raise_attribute_error
from ape.utils.misc import log_instead_of_fail


def valid_impl(api_class: Any) -> bool:
    """
    Check if an API class is valid. The class must not have any unimplemented
    abstract methods.

    Args:
        api_class (any)

    Returns:
        bool
    """

    if isinstance(api_class, tuple):
        return all(valid_impl(c) for c in api_class)

    # Is not an ABC base class or abstractdataclass
    if not hasattr(api_class, "__abstractmethods__"):
        return True  # not an abstract class

    return len(api_class.__abstractmethods__) == 0


def _get_unimplemented_methods_warning(api, plugin_name: str) -> str:
    unimplemented_methods: list[str] = []

    # Find the best API name to warn about.
    if isinstance(api, (list, tuple)):
        if classes := [p for p in api if hasattr(p, "__name__")]:
            # Likely only ever a single class in a registration, but just in case.
            api_name = " - ".join([p.__name__ for p in classes if hasattr(p, "__name__")])
            for api_cls in classes:
                unimplemented_methods.extend(_get_unimplemented_methods(api_cls))

        else:
            # This would only happen if the registration consisted of all primitives.
            api_name = " - ".join(api)

    elif hasattr(api, "__name__"):
        api_name = api.__name__
        unimplemented_methods.extend(_get_unimplemented_methods(api))

    else:
        api_name = api

    message = f"'{api_name}' from '{plugin_name}' is not fully implemented."
    if unimplemented_methods:
        # NOTE: Sorted for consistency.
        methods_str = ", ".join(sorted(unimplemented_methods))
        message = f"{message} Remaining abstract methods: '{methods_str}'."

    return message


def _get_unimplemented_methods(api) -> Iterable[str]:
    if (abstract_methods := getattr(api, "__abstractmethods__", None)) and hasattr(
        abstract_methods, "__iter__"
    ):
        return api.__abstractmethods__

    return []


class PluginManager:
    _unimplemented_plugins: list[str] = []

    def __init__(self) -> None:
        self.__registered = False

    @log_instead_of_fail(default="<PluginManager>")
    def __repr__(self) -> str:
        return f"<{PluginManager.__name__}>"

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Iterator[tuple[str, tuple]]:
        _assert_not_ipython_check(attr_name)

        # NOTE: The first time this method is called, the actual
        #  plugin registration occurs. Registration only happens once.
        self._register_plugins()

        if not hasattr(pluggy_manager.hook, attr_name):
            raise ApeAttributeError(f"{PluginManager.__name__} has no attribute '{attr_name}'.")

        # Do this to get access to the package name
        hook_fn = getattr(pluggy_manager.hook, attr_name)
        hookimpls = hook_fn.get_hookimpls()

        def get_plugin_name_and_hookfn(hook):
            return hook.plugin_name, getattr(hook.plugin, attr_name)()

        for plugin_name, results in map(get_plugin_name_and_hookfn, hookimpls):
            # NOTE: Some plugins return a tuple and some return iterators
            if not isinstance(results, Generator):
                validated_plugin = self._validate_plugin(plugin_name, results)
                if validated_plugin:
                    yield validated_plugin
            else:
                # Only if it's an iterator, provide results as a series
                for result in results:
                    validated_plugin = self._validate_plugin(plugin_name, result)
                    if validated_plugin:
                        yield validated_plugin

    @cached_property
    def registered_plugins(self) -> set[str]:
        plugins = list({n.replace("-", "_") for n in get_plugin_dists()})
        return {*plugins, *CORE_PLUGINS}

    def _register_plugins(self):
        if self.__registered:
            return

        handled = set()
        plugins = (n.replace("-", "_") for n in get_plugin_dists())
        for module_name in chain(plugins, iter(CORE_PLUGINS)):
            if module_name in handled:
                continue

            handled.add(module_name)
            try:
                module = import_module(module_name)
                pluggy_manager.register(module, name=module_name)
            except Exception as err:
                if module_name in CORE_PLUGINS or module_name == "ape":
                    # Always raise core plugin registration errors.
                    raise

                logger.warn_from_exception(err, f"Error loading plugin package '{module_name}'.")

        self.__registered = True

    def _validate_plugin(self, plugin_name: str, plugin_cls) -> tuple[str, tuple] | None:
        if valid_impl(plugin_cls):
            return clean_plugin_name(plugin_name), plugin_cls
        else:
            self._warn_not_fully_implemented_error(plugin_cls, plugin_name)
            return None

    def _warn_not_fully_implemented_error(self, results, plugin_name):
        if plugin_name in self._unimplemented_plugins:
            # Already warned
            return

        message = _get_unimplemented_methods_warning(results, plugin_name)
        logger.warning(message)

        # Record so we don't warn repeatedly
        self._unimplemented_plugins.append(plugin_name)



import json
import os
import shutil
from collections.abc import Iterable
from functools import cached_property
from pathlib import Path

import requests
from pydantic import model_validator

from ape.api.projects import DependencyAPI
from ape.exceptions import ProjectError
from ape.logging import logger
from ape.managers.project import _version_to_options
from ape.utils._github import _GithubClient, github_client
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.os import _remove_readonly, clean_path, extract_archive, get_package_path, in_tempdir


def _fetch_local(src: Path, destination: Path, config_override: dict | None = None):
    if src.is_dir():
        project = ManagerAccessMixin.Project(src, config_override=config_override)
        project.unpack(destination)
    elif src.is_file() and src.suffix == ".json":
        # Using a manifest directly as a dependency.
        if not destination.suffix:
            destination = destination / src.name

        destination.unlink(missing_ok=True)
        destination.parent.mkdir(parents=True, exist_ok=True)
        destination.write_text(src.read_text(), encoding="utf8")


class LocalDependency(DependencyAPI):
    """
    A dependency located on the local machine.
    """

    local: Path
    """
    The root path (and API defining key) to the dependency files.
    """

    version: str | None = None
    """
    Specified version.
    """

    @model_validator(mode="before")
    @classmethod
    def validate_local_path(cls, model):
        # Resolves the relative path so if the dependency API
        # data moves, it will still work.
        path = Path(model["local"]).expanduser()
        model["local"] = f"{path}"

        # Automatically include `"name"`.
        if "name" not in model:
            model["name"] = path.stem

        if path.is_absolute():
            return model

        elif "project" in model:
            # Just in case relative paths didn't get resolved.
            # Note: Generally, they should be resolved at model
            #  construction time, if parsing a config file normally.
            project = model.pop("project")
            model["local"] = (project / path).resolve()

        return model

    def __repr__(self) -> str:
        path = clean_path(self.local)
        return f"<LocalDependency local={path}, version={self.version_id}>"

    @property
    def package_id(self) -> str:
        path = self.local
        if in_tempdir(path):
            # Avoids never-ending tmp paths.
            return self.name

        else:
            return self.local.as_posix()

    @property
    def version_id(self) -> str:
        return self.version or "local"

    @property
    def uri(self) -> str:
        return self.local.as_uri()

    def fetch(self, destination: Path):
        if destination.is_dir():
            destination = destination / self.name

        _fetch_local(self.local, destination, config_override=self.config_override)


class GithubDependency(DependencyAPI):
    """
    A dependency from GitHub. Use the ``github`` key in your ``dependencies:``
    section of your ``ape-config.yaml`` file to declare a dependency from GitHub.

    Config example::

        dependencies:
          - name: OpenZeppelin
            github: OpenZeppelin/openzeppelin-contracts
            version: 4.4.0
    """

    github: str
    """
    The Github repo ID e.g. the organization name followed by the repo name,
    such as ``dapphub/erc20``.
    """

    ref: str | None = None
    """
    The branch or tag to use. When using this field
    instead of the 'release' field, the repository
    gets cloned instead of downloaded via the
    official GitHub release API.

    **NOTE**: Will be ignored if given a 'release'.
    """

    version: str | None = None
    """
    The release version to use. When using this
    field instead of the 'ref' field, the GitHub
    release API is used to fetch instead of cloning.

    **NOTE**: Will be ignored if given a 'ref'.
    """

    scheme: str = "https"
    """
    Either HTTPS or SSH; only used with `ref:`.
    """

    # Exists as property so can be changed for testing.
    _github_client: _GithubClient = github_client

    @model_validator(mode="before")
    @classmethod
    def _validate_model(cls, model):
        # branch -> ref
        if "branch" in model and "ref" not in model:
            # Handle branch as an alias.
            model["ref"] = model.pop("branch")

        if "name" not in model and "github" in model:
            # Calculate a default name.
            model["name"] = model["github"].split("/")[-1].lower()

        return model

    @model_validator(mode="after")
    def ensure_ref_or_version(self):
        if self.ref is None and self.version is None:
            raise ValueError("GitHub dependency must have either ref or version specified.")

        return self

    @property
    def package_id(self) -> str:
        return self.github

    @cached_property
    def version_id(self) -> str:
        if self.ref:
            return self.ref

        elif self.version and self.version != "latest":
            return self.version

        latest_release = self._github_client.get_latest_release(self.org_name, self.repo_name)
        return latest_release["tag_name"]

    @cached_property
    def org_name(self) -> str:
        return self.github.split("/")[0]

    @cached_property
    def repo_name(self) -> str:
        return self.github.split("/")[1]

    @property
    def uri(self) -> str:
        _uri = f"https://github.com/{self.github.strip('/')}"
        if self.version:
            version = f"v{self.version}" if not self.version.startswith("v") else self.version
            _uri = f"{_uri}/releases/tag/{version}"
        elif self.ref:
            _uri = f"{_uri}/tree/{self.ref}"

        return _uri

    def __repr__(self) -> str:
        cls_name = getattr(type(self), "__name__", GithubDependency.__name__)
        return f"<{cls_name} github={self.github}>"

    def fetch(self, destination: Path):
        destination.parent.mkdir(exist_ok=True, parents=True)
        if ref := self.ref:
            # NOTE: destination path should not exist at this point,
            #   so delete it in case it's left over from a failure.
            if destination.is_dir():
                shutil.rmtree(destination)

            # Fetch using git-clone approach (by git-reference).
            self._fetch_ref(ref, destination)
        else:
            # Fetch using Version API from GitHub.
            version = self.version or "latest"
            try:
                self._fetch_version(version, destination)
            except Exception as err_from_version_approach:
                logger.warning(
                    f"No official release found for version '{version}'. "
                    "Use `ref:` instead of `version:` for release tags. "
                    "Checking for matching tags..."
                )

                # NOTE: When using ref-from-a-version, ensure
                #   it didn't create the destination along the way;
                #   else, the ref is cloned in the wrong spot.
                if destination.is_dir():
                    shutil.rmtree(destination, onerror=_remove_readonly)
                try:
                    self._fetch_ref(version, destination)
                except Exception:
                    # NOTE: Ignore this error, it was merely a last attempt.
                    raise err_from_version_approach

    def _fetch_ref(self, ref: str, destination: Path):
        options = _version_to_options(ref)
        attempt = 0
        num_attempts = len(options)
        for ref in options:
            attempt += 1
            try:
                self._github_client.clone_repo(
                    self.org_name,
                    self.repo_name,
                    destination,
                    branch=ref,
                    scheme=self.scheme,
                )
            except Exception:
                if attempt == num_attempts:
                    raise  # This error!

                # Try another option.
                continue

            else:
                # Was successful! Don't try anymore.
                break

    def _fetch_version(self, version: str, destination: Path):
        destination.mkdir(parents=True, exist_ok=True)
        options = _version_to_options(version)
        attempt = 0
        max_attempts = len(options)

        for vers in options:
            attempt += 1
            try:
                self._github_client.download_package(
                    self.org_name, self.repo_name, vers, destination
                )
            except Exception:
                if attempt == max_attempts:
                    raise  # This error!

                # Try another option.
                continue

            else:
                # Was successful! Don't try anymore.
                break


class NpmDependency(DependencyAPI):
    """
    A dependency from the Node Package Manager (NPM).

    Config example::

        dependencies:
          - name: safe-singleton-factory
            npm: "@gnosis.pm/safe-singleton-factory"
            version: 1.0.14
    """

    npm: Path
    """
    The NPM repo ID e.g. the organization name followed by the repo name,
    such as ``"@gnosis.pm/safe-singleton-factory"``.
    Note: Resolves to a 'path' after serialization.
    The package must already be installed!
    """

    version: str | None = None
    """
    Specify the version, if not wanting to use discovered version
    from install.
    """

    @model_validator(mode="before")
    @classmethod
    def validate_local_npm(cls, model):
        # Resolves the relative path so if the dependency API
        # data moves, it will still work.
        npm = Path(model["npm"])
        if npm.is_absolute():
            return model

        elif "project" in model:
            # Just in case relative paths didn't get resolved.
            # Note: Generally, they should be resolved at model
            #  construction time, if parsing a config file normally.
            project = model.pop("project")
            if isinstance(project, str):
                project_path = Path(project)
            elif isinstance(project, Path):
                project_path = project
            else:
                project_path = project.path

            for base_path in (project_path, Path.home()):
                path = base_path / "node_modules" / npm
                if path.is_dir():
                    model["npm"] = npm = path
                    break

        return model

    @cached_property
    def version_id(self) -> str:
        if version := (
            self.version
            or self.version_from_installed_package_json
            or self.version_from_project_package_json
        ):
            return version

        raise ProjectError(
            f"Missing version for NPM dependency '{self.name}'. Have you run `npm install`?"
        )

    @property
    def package_id(self) -> str:
        return str(self.npm).split("node_modules")[-1].strip(os.path.sep)

    @cached_property
    def version_from_installed_package_json(self) -> str | None:
        """
        The version from package.json in the installed package.
        Requires having run `npm install`.
        """
        return _get_version_from_package_json(self.npm)

    @cached_property
    def version_from_project_package_json(self) -> str | None:
        """
        The version from your project's package.json, if exists.
        """
        return _get_version_from_package_json(
            self.local_project.path, dict_path=("dependencies", self.package_id)
        )

    @property
    def uri(self) -> str:
        return f"https://www.npmjs.com/package/{self.npm}/v/{self.version_id}"

    def fetch(self, destination: Path):
        if self.npm.is_dir():
            if destination.is_dir():
                destination = destination / self.name

            shutil.copytree(self.npm, destination)
        else:
            raise ProjectError(f"NPM package '{self.package_id}' not installed.")


def _get_version_from_package_json(
    base_path: Path, dict_path: Iterable[str | Path] | None = None
) -> str | None:
    package_json = base_path / "package.json"
    if not package_json.is_file():
        return None

    try:
        data = json.loads(package_json.read_text())
    except Exception as err:
        logger.warning(f"Failed to parse package.json: {err}")
        return None

    for key in dict_path or []:
        if key not in data:
            return None

        data = data[key]

    if isinstance(data, str):
        return data

    elif not isinstance(data, dict):
        return None

    return data.get("version")


class PythonDependency(DependencyAPI):
    """
    A dependency installed from Python tooling, such as `pip`.
    """

    site_package: str | None = None
    """
    The Python site-package name, such as ``"snekmate"``. Cannot use
    with ``pypi:``. Requires the dependency to have been installed
    either via ``pip`` or something alike.
    """

    pypi: str | None = None
    """
    The ``pypi`` reference, such as ``"snekmate"``. Cannot use with
    ``python:``. When set, downloads the dependency from ``pypi``
    using HTTP directly (not ``pip``).
    """

    version: str | None = None
    """
    Optionally specify the version expected to be installed.
    """

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, values):
        if "python" in values:
            # `.python` is the old key but we have to always support it
            # so dependencies-of-dependencies always work, even when referencing
            # older projects.
            values["site_package"] = values.pop("python")

        if "name" not in values:
            if name := values.get("site_package") or values.get("pypi"):
                values["name"] = name
            else:
                raise ValueError(
                    "Must set either 'pypi:' or 'site_package': when using Python dependencies"
                )

        return values

    @cached_property
    def path(self) -> Path | None:
        if self.pypi:
            # Is pypi: specified; has no special path.
            return None

        elif python := self.site_package:
            try:
                return get_package_path(python)
            except ValueError as err:
                raise ProjectError(str(err)) from err

        return None

    @property
    def package_id(self) -> str:
        if pkg_id := (self.pypi or self.site_package):
            return pkg_id

        raise ProjectError("Must provide either 'pypi:' or 'python:' for python-base dependencies.")

    @property
    def python(self) -> str | None:
        # For backwards-compat; serves as an undocumented alias.
        return self.site_package

    @property
    def version_id(self) -> str:
        if self.version:
            # It is helpful to just return the cfg version here
            # so uninstalled dependencies can attempt to set up their caches.
            return self.version

        elif self.pypi:
            # Version available in package data.
            vers = self._get_version_from_package_data()

        elif self.site_package:
            # Python dependency not installed; attempt to use latest from pypi.
            if pkg_vers := self.version_from_package_data:
                return pkg_vers

            # Force the user to specify the version, as it is not installed and not
            # available on PyPI.
            raise ProjectError(
                f"Dependency '{self.name}' not installed. Either install or specify the `version:` to continue."
            )

        else:
            raise ProjectError(
                "Must provide either 'pypi:' or 'python:' for python-base dependencies."
            )

        return vers

    @property
    def uri(self) -> str:
        if self.pypi:
            return self.download_archive_url

        elif self.site_package and (path := self.path):
            # Local site-package path.
            return path.as_uri()

        else:
            raise ProjectError(
                "Must provide either 'pypi:' or 'python:' for python-base dependencies."
            )

    @cached_property
    def package_data(self) -> dict:
        url = f"https://pypi.org/pypi/{self.package_id}/json"
        response = requests.get(url)

        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                # There is no available package data on PyPI; use empty dict.
                return {}

            else:
                # It should have worked in this case, so it is good to raise an error.
                raise ProjectError(
                    f"Problem downloading package data for '{self.package_id}': {err}"
                )

        return response.json()

    @cached_property
    def version_from_package_data(self) -> str | None:
        return self.package_data.get("info", {}).get("version")

    @cached_property
    def download_archive_url(self) -> str:
        if not (version := self.version):
            if not (version := self.version_from_package_data):
                # Not sure if this is possible, but just in case API data changes or something.
                raise ProjectError(f"Unable to find version for package '{self.package_id}'.")

        releases = self.package_data.get("releases", {})
        if version not in releases:
            raise ProjectError(f"Version '{version}' not found for package '{self.package_id}'.")

        # Find the first zip file in the specified release.
        for file_info in releases[version]:
            if file_info.get("packagetype") != "sdist":
                continue

            return file_info["url"]

        raise ProjectError(
            f"No zip file found for package '{self.package_id}' with version '{version}' on PyPI."
        )

    def fetch(self, destination: Path):
        if self.pypi:
            self._fetch_from_pypi(destination)
        elif path := self.path:
            # 'python:' key.
            _fetch_local(path, destination, config_override=self.config_override)

    def _fetch_from_pypi(self, destination: Path):
        archive_path = self._fetch_archive_file(destination)
        extract_archive(archive_path)
        archive_path.unlink(missing_ok=True)

    def _fetch_archive_file(self, destination) -> Path:
        logger.info(f"Fetching python dependency '{self.package_id}' from 'pypi.")
        download_url = self.download_archive_url
        filename = download_url.split("/")[-1]
        destination.mkdir(exist_ok=True, parents=True)
        archive_destination = destination / filename
        with requests.get(download_url, stream=True) as response:
            response.raise_for_status()
            with open(archive_destination, "wb") as file:
                for chunk in response.iter_content(chunk_size=8192):  # 8 KB
                    file.write(chunk)

        return archive_destination

    def _get_version_from_package_data(self) -> str:
        if vers := self.version_from_package_data:
            return vers

        # I doubt this is a possible condition, but just in case.
        raise ProjectError(f"Missing version from PyPI for package '{self.package_id}'.")
